import freeOrionAIInterface as fo
from logging import debug, warning

import AIstate
import CombatRatingsAI
import EspionageAI
import FleetUtilsAI
import InvasionAI
import MilitaryAI
import MoveUtilsAI
from AIDependencies import INVALID_ID
from aistate_interface import get_aistate
from common.fo_typing import FleetId
from EnumsAI import MissionType
from fleet_orders import (
    AIFleetOrder,
    OrderColonize,
    OrderDefend,
    OrderInvade,
    OrderMilitary,
    OrderMove,
    OrderOutpost,
    OrderPause,
)
from freeorion_tools import (
    assertion_fails,
    combine_ratings,
    get_fleet_position,
    get_partial_visibility_turn,
)
from target import Target, TargetFleet, TargetPlanet, TargetSystem
from universe.system_network import get_neighbors

ORDERS_FOR_MISSION = {
    MissionType.EXPLORATION: OrderPause,
    MissionType.OUTPOST: OrderOutpost,
    MissionType.COLONISATION: OrderColonize,
    MissionType.INVASION: OrderInvade,
    MissionType.MILITARY: OrderMilitary,
    # SECURE is mostly same as MILITARY, but waits for system removal from all targeted system lists
    # (invasion, colonization, outpost, blockade) before clearing
    MissionType.SECURE: OrderMilitary,
    MissionType.PROTECT_REGION: OrderPause,
    MissionType.ORBITAL_INVASION: OrderInvade,
    MissionType.ORBITAL_OUTPOST: OrderOutpost,
    MissionType.ORBITAL_DEFENSE: OrderDefend,
}

COMBAT_MISSION_TYPES = (
    MissionType.MILITARY,
    MissionType.SECURE,
    MissionType.PROTECT_REGION,
)

MERGEABLE_MISSION_TYPES = (
    MissionType.MILITARY,
    MissionType.INVASION,
    MissionType.PROTECT_REGION,
    MissionType.ORBITAL_INVASION,
    MissionType.SECURE,
    MissionType.ORBITAL_DEFENSE,
)


class AIFleetMission:
    """
    Stores information about AI mission. Every mission has fleetID and AI targets depending upon AI fleet mission type.
    :type orders: list[AIFleetOrder]
    :type target: target.Target | None
    """

    def __init__(self, fleet_id: FleetId):
        self.orders = []
        self.fleet = TargetFleet(fleet_id)
        self.type = None
        self.target = None

    def __setstate__(self, state):
        target_type = state.pop("target_type")
        if state["target"] is not None:
            object_map = {
                TargetPlanet.object_name: TargetPlanet,
                TargetSystem.object_name: TargetSystem,
                TargetFleet.object_name: TargetFleet,
            }
            state["target"] = object_map[target_type](state["target"])

        state["fleet"] = TargetFleet(state["fleet"])
        self.__dict__ = state

    def __getstate__(self):
        retval = dict(self.__dict__)
        # do only store the fleet id not the Fleet object
        retval["fleet"] = self.fleet.id

        # store target type and id rather than the object
        if self.target is not None:
            retval["target_type"] = self.target.object_name
            retval["target"] = self.target.id
        else:
            retval["target_type"] = None
            retval["target"] = None

        return retval

    def set_target(self, mission_type: MissionType, target: Target):
        """
        Set mission and target for this fleet.
        """
        if self.type == mission_type and self.target == target:
            return
        if self.type or self.target:
            debug(f"{self.fleet}: change mission assignment from {self.type}:{self.target} to {mission_type}:{target}")
        self.type = mission_type
        self.target = target

    def clear_target(self):
        """Clear target and mission for this fleet."""
        self.target = None
        self.type = None

    def has_target(self, mission_type: MissionType, target: Target) -> bool:
        """
        Check if fleet has specified mission_type and target.
        """
        return self.type == mission_type and self.target == target

    def clear_fleet_orders(self):
        """Clear this fleets orders but do not clear mission and target."""
        self.orders = []

    def _get_fleet_order_from_target(self, mission_type: MissionType, target: Target) -> AIFleetOrder:
        """
        Get a fleet order according to mission type and target.
        """
        fleet_target = TargetFleet(self.fleet.id)
        return ORDERS_FOR_MISSION[mission_type](fleet_target, target)

    def check_mergers(self, context: str = ""):
        """
        Merge local fleets with same mission into this fleet.

        :param context: Context of the function call for logging purposes
        """
        debug(f"Considering to merge {self}")
        if self.type not in MERGEABLE_MISSION_TYPES:
            debug("Mission type does not allow merging.")
            return

        if not self.target:
            debug("Mission has no valid target - do not merge.")
            return

        universe = fo.getUniverse()
        empire_id = fo.empireID()

        fleet_id = self.fleet.id
        main_fleet = universe.getFleet(fleet_id)
        main_fleet_system_id = main_fleet.systemID
        if main_fleet_system_id == INVALID_ID:
            debug("Can't merge: fleet in middle of starlane.")
            return

        # only merge PROTECT_REGION if there is any threat near target
        if self.type == MissionType.PROTECT_REGION:
            neighbor_systems = get_neighbors(self.target.id)
            if not any(MilitaryAI.get_system_local_threat(sys_id) for sys_id in neighbor_systems):
                debug("Not merging PROTECT_REGION fleet - no threat nearby.")
                return

        destroyed_list = set(universe.destroyedObjectIDs(empire_id))
        aistate = get_aistate()
        system_status = aistate.systemStatus[main_fleet_system_id]
        other_fleets_here = [
            fid
            for fid in system_status.get("myFleetsAccessible", [])
            if fid != fleet_id and fid not in destroyed_list and universe.getFleet(fid).ownedBy(empire_id)
        ]
        if not other_fleets_here:
            debug("No other fleets here")
            return

        for fid in other_fleets_here:
            fleet_mission = aistate.get_fleet_mission(fid)
            if fleet_mission.type != self.type or fleet_mission.target != self.target:
                debug("Local candidate %s does not have same mission." % fleet_mission)
                continue
            FleetUtilsAI.merge_fleet_a_into_b(fid, fleet_id, context=f"Order {context} of mission {self}")

    def _is_valid_fleet_mission_target(self, mission_type: MissionType, target: Target):  # noqa: C901
        if not target:
            return False
        if mission_type == MissionType.EXPLORATION:
            if isinstance(target, TargetSystem):
                empire = fo.getEmpire()
                if not empire.hasExploredSystem(target.id):
                    return True
        elif mission_type in [MissionType.OUTPOST, MissionType.ORBITAL_OUTPOST]:
            fleet = self.fleet.get_object()
            if not fleet.hasOutpostShips:
                return False
            if isinstance(target, TargetPlanet):
                planet = target.get_object()
                if planet.unowned:
                    return True
        elif mission_type == MissionType.COLONISATION:
            fleet = self.fleet.get_object()
            if not fleet.hasColonyShips:
                return False
            if isinstance(target, TargetPlanet):
                planet = target.get_object()
                population = planet.initialMeterValue(fo.meterType.population)
                if planet.unowned or (planet.owner == fleet.owner and population == 0):
                    return True
        elif mission_type in [MissionType.INVASION, MissionType.ORBITAL_INVASION]:
            fleet = self.fleet.get_object()
            if not fleet.hasTroopShips:
                return False
            if isinstance(target, TargetPlanet):
                planet = target.get_object()
                # TODO remove latter portion of next check in light of invasion retargeting, or else correct logic
                if not planet.unowned or planet.owner != fleet.owner:
                    return True
        elif mission_type in [
            MissionType.MILITARY,
            MissionType.SECURE,
            MissionType.ORBITAL_DEFENSE,
            MissionType.PROTECT_REGION,
        ]:
            if isinstance(target, TargetSystem):
                return True
        # TODO: implement other mission types
        return False

    def clean_invalid_targets(self):
        """clean invalid AITargets"""
        if not self._is_valid_fleet_mission_target(self.type, self.target):
            self.target = None
            self.type = None

    def _check_abort_mission(self, fleet_order: AIFleetOrder):  # noqa: C901
        """checks if current mission (targeting a planet) should be aborted"""
        planet_stealthed = False
        target_is_planet = fleet_order.target and isinstance(fleet_order.target, TargetPlanet)
        planet = None
        if target_is_planet:
            planet = fleet_order.target.get_object()
            # Check visibility prediction, but if somehow still have current visibility, don't
            # abort the mission yet
            if not EspionageAI.colony_detectable_by_empire(planet.id, empire=fo.empireID()):
                if get_partial_visibility_turn(planet.id) == fo.currentTurn():
                    debug(
                        "EspionageAI predicts planet id %d to be stealthed" % planet.id
                        + ", but somehow have current visibity anyway, so won't trigger mission abort"
                    )
                else:
                    debug("EspionageAI predicts we can no longer detect %s, will abort mission" % fleet_order.target)
                    planet_stealthed = True
        if target_is_planet and not planet_stealthed:
            # TBD use fleet_order.is_valid or add fleet_order.should_abort
            if isinstance(fleet_order, OrderColonize):
                if planet.initialMeterValue(fo.meterType.population) == 0 and (
                    planet.ownedBy(fo.empireID()) or planet.unowned
                ):
                    return False
            elif isinstance(fleet_order, OrderOutpost):
                if planet.unowned:
                    return False
            elif isinstance(fleet_order, OrderInvade):
                if planet.owner == fo.getEmpire().empireID:
                    debug(f"Abandoning invasion for {planet}, it already belongs to us")
                    return True
                if planet.currentMeterValue(fo.meterType.maxShield):
                    military_support_fleets = MilitaryAI.get_military_fleets_with_target_system(planet.systemID)
                    if not military_support_fleets:
                        # Maybe try again later...
                        debug(
                            f"Abandoning Invasion mission for {planet} because target has nonzero max shields "
                            f"and there is no military fleet assigned to secure the target system."
                        )
                        return True
                return False
            else:
                return False

        # canceling fleet orders
        debug("   %s" % fleet_order)
        debug("Fleet %d had a target planet that is no longer valid for this mission; aborting." % self.fleet.id)
        FleetUtilsAI.split_fleet(self.fleet.id)
        return True

    def _check_retarget_invasion(self):  # noqa: C901
        """checks if an invasion mission should be retargeted"""
        universe = fo.getUniverse()
        empire_id = fo.empireID()
        fleet_id = self.fleet.id
        fleet = universe.getFleet(fleet_id)
        if fleet.systemID == INVALID_ID:
            # next_loc = fleet.nextSystemID
            return  # TODO: still check
        system = universe.getSystem(fleet.systemID)
        if not system:
            return
        orders = self.orders
        last_sys_target = INVALID_ID
        if orders:
            last_sys_target = orders[-1].target.id
        if last_sys_target == fleet.systemID:
            return  # TODO: check for best local target
        open_targets = []
        already_targeted = InvasionAI.get_invasion_targeted_planet_ids(system.planetIDs, MissionType.INVASION)
        aistate = get_aistate()
        for pid in system.planetIDs:
            if pid in already_targeted or (pid in aistate.qualifyingTroopBaseTargets):
                continue
            planet = universe.getPlanet(pid)
            if planet.unowned or (planet.owner == empire_id):
                continue
            if (planet.initialMeterValue(fo.meterType.shield)) <= 0:
                open_targets.append(pid)
        if not open_targets:
            return
        troops_in_fleet = FleetUtilsAI.count_troops_in_fleet(fleet_id)
        target_id = INVALID_ID
        best_score = -1
        target_troops = 0
        #
        for pid, rating in InvasionAI.assign_invasion_values(open_targets).items():
            p_score, p_troops = rating
            if p_score > best_score:
                if p_troops >= troops_in_fleet:
                    continue
                best_score = p_score
                target_id = pid
                target_troops = p_troops
        if target_id == INVALID_ID:
            return

        debug("\t Splitting and retargetting fleet %d" % fleet_id)
        new_fleets = FleetUtilsAI.split_fleet(fleet_id)
        self.clear_target()  # TODO: clear from foAIstate
        self.clear_fleet_orders()
        troops_needed = max(0, target_troops - FleetUtilsAI.count_troops_in_fleet(fleet_id))
        min_stats = {"troopCapacity": troops_needed}
        target_stats = {"troopCapacity": troops_needed}
        found_fleets = []
        # TODO check if next statement does not mutate any global states and can be removed

        _ = FleetUtilsAI.get_fleets_for_mission(
            target_stats,
            min_stats,
            {},
            starting_system=fleet.systemID,  # noqa: F841
            fleet_pool_set=set(new_fleets),
            fleet_list=found_fleets,
        )
        for fid in found_fleets:
            FleetUtilsAI.merge_fleet_a_into_b(fid, fleet_id)
        target = TargetPlanet(target_id)
        self.set_target(MissionType.INVASION, target)
        self.generate_fleet_orders()

    def need_to_pause_movement(self, last_move_target_id: int, new_move_order: OrderMove) -> bool:
        """
        When a fleet has consecutive move orders, assesses whether something about the interim destination warrants
        forcing a stop (such as a military fleet choosing to engage with an enemy fleet about to enter the same system,
        or it may provide a good vantage point to view current status of next system in path). Assessments about whether
        the new destination is suitable to move to are (currently) separately made by OrderMove.can_issue_order()
        :param last_move_target_id:
        :param new_move_order:
        """
        fleet = self.fleet.get_object()
        # don't try skipping over more than one System
        if fleet.nextSystemID != last_move_target_id:
            return True
        universe = fo.getUniverse()
        current_dest_system = universe.getSystem(fleet.nextSystemID)
        if not current_dest_system:
            # shouldn't really happen, but just to be safe
            return True
        distance_to_next_system = (
            (fleet.x - current_dest_system.x) ** 2 + (fleet.y - current_dest_system.y) ** 2
        ) ** 0.5
        surplus_travel_distance = fleet.speed - distance_to_next_system
        # if need more than one turn to reach current destination, then don't add another jump yet
        if surplus_travel_distance < 0:
            return True
        # TODO: add assessments for other situations we'd prefer to pause, such as cited above re military fleets, and
        # for situations where high value fleets like colony fleets might deem it safest to stop and look around
        # before proceeding
        return False

    def issue_fleet_orders(self):  # noqa: C901
        """issues AIFleetOrders which can be issued in system and moves to next one if is possible"""
        # TODO: priority
        order_completed = True

        debug(
            "\nChecking orders for fleet %s (on turn %d), with mission type %s and target %s",
            self.fleet.get_object(),
            fo.currentTurn(),
            self.type or "No mission",
            self.target or "No Target",
        )
        if MissionType.INVASION == self.type:
            self._check_retarget_invasion()
        just_issued_move_order = False
        last_move_target_id = INVALID_ID
        for fleet_order in self.orders:
            if isinstance(fleet_order, (OrderColonize, OrderOutpost, OrderInvade)) and self._check_abort_mission(
                fleet_order
            ):
                self.clear_fleet_orders()
                self.clear_target()
                return
        aistate = get_aistate()
        for fleet_order in self.orders:
            if just_issued_move_order and self.fleet.get_object().systemID != last_move_target_id:
                # having just issued a move order, we will normally stop issuing orders this turn, except that if there
                # are consecutive move orders we will consider moving through the first destination rather than stopping
                # Without the below noinspection directive, PyCharm is concerned about the 2nd part of the test
                # noinspection PyTypeChecker
                if not isinstance(fleet_order, OrderMove) or self.need_to_pause_movement(
                    last_move_target_id, fleet_order
                ):
                    break
            debug("Checking order: %s" % fleet_order)
            self.check_mergers(context=str(fleet_order))
            if fleet_order.can_issue_order(verbose=False):
                # only move if all other orders completed
                if isinstance(fleet_order, OrderMove) and order_completed:
                    debug("Issuing fleet order %s" % fleet_order)
                    fleet_order.issue_order()
                    just_issued_move_order = True
                    last_move_target_id = fleet_order.target.id
                elif not isinstance(fleet_order, OrderMove):
                    debug("Issuing fleet order %s" % fleet_order)
                    fleet_order.issue_order()
                else:
                    debug("NOT issuing (even though can_issue) fleet order %s" % fleet_order)
                issued = "issued" if fleet_order.order_issued else "not issued"
                executed = "fully executed" if fleet_order.executed else "not fully executed"
                debug(f"Order {issued} issued and {executed}.")
                if not fleet_order.executed:
                    order_completed = False
            else:  # check that we're not held up by a Big Monster
                if fleet_order.order_issued:
                    # A previously issued order that wasn't instantly executed must have had cirumstances change so that
                    # the order can't currently be reissued (or perhaps simply a savegame has been reloaded on the same
                    # turn the order was issued).
                    if not fleet_order.executed:
                        order_completed = False
                    # Go on to the next order.
                    continue
                debug("CAN'T issue fleet order %s because:" % fleet_order)
                fleet_order.can_issue_order(verbose=True)
                if isinstance(fleet_order, OrderMove):
                    this_system_id = fleet_order.target.id
                    this_status = aistate.systemStatus.setdefault(this_system_id, {})
                    threat_threshold = fo.currentTurn() * MilitaryAI.cur_best_mil_ship_rating() / 4.0
                    if this_status.get("monsterThreat", 0) > threat_threshold:
                        # if this move order is not this mil fleet's final destination, and blocked by Big Monster,
                        # release and hope for more effective reassignment
                        if (
                            self.type not in (MissionType.MILITARY, MissionType.SECURE)
                            or fleet_order != self.orders[-1]
                        ):
                            debug(
                                "Aborting mission due to being blocked by Big Monster at system %d, threat %d"
                                % (this_system_id, aistate.systemStatus[this_system_id]["monsterThreat"])
                            )
                            debug("Full set of orders were:")
                            for this_order in self.orders:
                                debug(" - %s" % this_order)
                            self.clear_fleet_orders()
                            self.clear_target()
                            return
                break  # do not order the next order until this one is finished.
        else:  # went through entire order list
            if order_completed:
                debug("Final order is completed")
                orders = self.orders
                last_order = orders[-1] if orders else None
                universe = fo.getUniverse()

                if last_order and isinstance(last_order, OrderColonize):
                    planet = universe.getPlanet(last_order.target.id)
                    sys_partial_vis_turn = get_partial_visibility_turn(planet.systemID)
                    planet_partial_vis_turn = get_partial_visibility_turn(planet.id)
                    if planet_partial_vis_turn == sys_partial_vis_turn and not planet.initialMeterValue(
                        fo.meterType.population
                    ):
                        warning(
                            "Fleet %s has tentatively completed its "
                            "colonize mission but will wait to confirm population.",
                            self.fleet,
                        )
                        debug("    Order details are %s" % last_order)
                        debug(
                            f"    Order is valid: {last_order.is_valid()}; issued: {last_order.order_issued}; executed: {last_order.executed}"
                        )
                        if not last_order.is_valid():
                            source_target = last_order.fleet
                            target_target = last_order.target
                            debug(
                                f"        source target validity: {bool(source_target)}; target target validity: {bool(target_target)} "
                            )
                        return  # colonize order must not have completed yet
                clear_all = True
                last_sys_target = INVALID_ID
                secure_targets = set(
                    AIstate.colonyTargetedSystemIDs
                    + AIstate.outpostTargetedSystemIDs
                    + AIstate.invasionTargetedSystemIDs
                )
                if last_order and isinstance(last_order, OrderMilitary):
                    last_sys_target = last_order.target.id
                    # not doing this until decide a way to release from a SECURE mission
                    # if (MissionType.SECURE == self.type) or
                    if last_sys_target in secure_targets:  # consider a secure mission
                        if last_sys_target in AIstate.colonyTargetedSystemIDs:
                            secure_type = "Colony"
                        elif last_sys_target in AIstate.outpostTargetedSystemIDs:
                            secure_type = "Outpost"
                        elif last_sys_target in AIstate.invasionTargetedSystemIDs:
                            secure_type = "Invasion"
                        else:
                            secure_type = "Unidentified"
                        debug(
                            "Fleet %d has completed initial stage of its mission "
                            "to secure system %d (targeted for %s), "
                            "may release a portion of ships" % (self.fleet.id, last_sys_target, secure_type)
                        )
                        clear_all = False

                # for PROTECT_REGION missions, only release fleet if no more threat
                if self.type == MissionType.PROTECT_REGION:
                    # use military logic code below to determine if can release
                    # any or even all of the ships.
                    clear_all = False
                    last_sys_target = self.target.id
                    debug("Check if PROTECT_REGION mission with target %d is finished.", last_sys_target)

                # for SECURE missions, only release fleet if no settle or invade mission is ongoing
                if self.type == MissionType.SECURE and self.target.id in secure_targets:
                    debug(f"Secure mission for {self.target} not yet finished")
                    return

                fleet_id = self.fleet.id
                if clear_all:
                    if orders:
                        debug("Fleet %d has completed its mission; clearing all orders and targets." % self.fleet.id)
                        debug("Full set of orders were:")
                        for this_order in orders:
                            debug("\t\t %s" % this_order)
                        self.clear_fleet_orders()
                        self.clear_target()
                        if aistate.get_fleet_role(fleet_id) in (MissionType.MILITARY, MissionType.SECURE):
                            allocations = MilitaryAI.get_military_fleets(
                                mil_fleets_ids=[fleet_id], try_reset=False, thisround="Fleet %d Reassignment" % fleet_id
                            )
                            if allocations:
                                MilitaryAI.assign_military_fleets_to_systems(
                                    use_fleet_id_list=[fleet_id], allocations=allocations
                                )
                    else:  # no orders
                        debug("No Current Orders")
                else:
                    potential_threat = combine_ratings(
                        MilitaryAI.get_system_local_threat(last_sys_target),
                        MilitaryAI.get_system_neighbor_threat(last_sys_target),
                    )
                    threat_present = potential_threat > 0
                    debug("Fleet threat present? %s", threat_present)
                    target_system = universe.getSystem(last_sys_target)
                    if not threat_present and target_system:
                        for pid in target_system.planetIDs:
                            planet = universe.getPlanet(pid)
                            if (
                                planet
                                and planet.owner != fo.empireID()
                                and planet.currentMeterValue(fo.meterType.maxDefense) > 0
                            ):
                                debug("Found local planetary threat: %s", planet)
                                threat_present = True
                                break
                    if not threat_present:
                        debug("No current threat in target system; releasing a portion of ships.")
                        # at least first stage of current task is done;
                        # release extra ships for potential other deployments
                        new_fleets = FleetUtilsAI.split_fleet(self.fleet.id)
                        if self.type == MissionType.PROTECT_REGION:
                            self.clear_fleet_orders()
                            self.clear_target()
                            new_fleets.append(self.fleet.id)
                    else:
                        debug("Threat remains in target system; Considering to release some ships.")
                        new_fleets = []
                        fleet_portion_to_remain = self._portion_of_fleet_needed_here()
                        if fleet_portion_to_remain >= 1:
                            debug("Can not release fleet yet due to large threat.")
                        elif fleet_portion_to_remain > 0:
                            debug("Not all ships are needed here - considering releasing a few")
                            # TODO: Rate against specific enemy threat cause
                            fleet_remaining_rating = CombatRatingsAI.get_fleet_rating(fleet_id)
                            fleet_min_rating = fleet_portion_to_remain * fleet_remaining_rating
                            debug(
                                "Starting rating: %.1f, Target rating: %.1f", fleet_remaining_rating, fleet_min_rating
                            )
                            allowance = CombatRatingsAI.rating_needed(fleet_remaining_rating, fleet_min_rating)
                            debug("May release ships with total rating of %.1f", allowance)
                            for ship_id in self.fleet.get_object().shipIDs:
                                if len(self.fleet.get_object().shipIDs) == 1:
                                    break
                                ship_rating = CombatRatingsAI.get_ship_rating(ship_id)
                                debug("Considering to release ship %d with rating %.1f", ship_id, ship_rating)
                                if ship_rating > allowance:
                                    debug("Remaining rating insufficient. Not released.")
                                    continue
                                debug("Splitting from fleet.")
                                new_fleet_id = FleetUtilsAI.split_ship_from_fleet(fleet_id, ship_id)
                                if assertion_fails(new_fleet_id != INVALID_ID):
                                    break
                                new_fleets.append(new_fleet_id)
                                fleet_remaining_rating = CombatRatingsAI.rating_difference(
                                    fleet_remaining_rating, ship_rating
                                )
                                allowance = CombatRatingsAI.rating_difference(fleet_remaining_rating, fleet_min_rating)
                                debug(
                                    "Remaining fleet rating: %.1f - Allowance: %.1f", fleet_remaining_rating, allowance
                                )
                            if new_fleets:
                                aistate.get_fleet_role(fleet_id, force_new=True)
                                aistate.update_fleet_rating(fleet_id)
                                aistate.ensure_have_fleet_missions(new_fleets)
                        else:
                            debug("Planetary defenses are deemed sufficient. Release fleet.")
                            new_fleets = FleetUtilsAI.split_fleet(self.fleet.id)

                    new_military_fleets = []
                    for fleet_id in new_fleets:
                        if aistate.get_fleet_role(fleet_id) in COMBAT_MISSION_TYPES:
                            new_military_fleets.append(fleet_id)
                    allocations = []
                    if new_military_fleets:
                        allocations = MilitaryAI.get_military_fleets(
                            mil_fleets_ids=new_military_fleets,
                            try_reset=False,
                            thisround="Fleet Reassignment %s" % new_military_fleets,
                        )
                    if allocations:
                        MilitaryAI.assign_military_fleets_to_systems(
                            use_fleet_id_list=new_military_fleets, allocations=allocations
                        )

    def _portion_of_fleet_needed_here(self):
        """Calculate the portion of the fleet needed in target system considering enemy forces."""
        # TODO check rating against planets
        if assertion_fails(self.type in COMBAT_MISSION_TYPES, msg=str(self)):
            return 0
        if assertion_fails(self.target and self.target.id != INVALID_ID, msg=str(self)):
            return 0
        system_id = self.target.id
        aistate = get_aistate()
        local_defenses = MilitaryAI.get_my_defense_rating_in_system(system_id)
        potential_threat = combine_ratings(
            MilitaryAI.get_system_local_threat(system_id), MilitaryAI.get_system_neighbor_threat(system_id)
        )
        universe = fo.getUniverse()
        system = universe.getSystem(system_id)

        # tally planetary defenses
        total_defense = total_shields = 0
        for planet_id in system.planetIDs:
            planet = universe.getPlanet(planet_id)
            total_defense += planet.currentMeterValue(fo.meterType.defense)
            total_shields += planet.currentMeterValue(fo.meterType.shield)
        planetary_ratings = total_defense * (total_shields + total_defense)
        potential_threat += planetary_ratings  # TODO: rewrite to return min rating vs planets as well

        # consider safety factor just once here rather than everywhere below
        safety_factor = aistate.character.military_safety_factor()
        potential_threat *= safety_factor

        # TODO: Rate against specific threat here
        fleet_rating = CombatRatingsAI.get_fleet_rating(self.fleet.id)
        return CombatRatingsAI.rating_needed(potential_threat, local_defenses) / float(max(fleet_rating, 1.0))

    def generate_fleet_orders(self):
        """generates AIFleetOrders from fleets targets to accomplish"""
        universe = fo.getUniverse()
        fleet_id = self.fleet.id
        fleet = universe.getFleet(fleet_id)
        if (not fleet) or fleet.empty or (fleet_id in universe.destroyedObjectIDs(fo.empireID())):
            # fleet was probably merged into another or was destroyed
            get_aistate().delete_fleet_info(fleet_id)
            return

        # TODO: priority
        self.clear_fleet_orders()
        system_id = fleet.systemID
        # if fleet doesn't have any mission,
        # then repair if needed or resupply if is current location not in supplyable system
        empire = fo.getEmpire()
        fleet_supplyable_system_ids = empire.fleetSupplyableSystemIDs
        # if (not self.hasAnyAIMissionTypes()):
        if not self.target and (
            system_id
            not in set(
                AIstate.colonyTargetedSystemIDs + AIstate.outpostTargetedSystemIDs + AIstate.invasionTargetedSystemIDs
            )
        ):
            if self._need_repair():
                repair_fleet_order = MoveUtilsAI.get_repair_fleet_order(self.fleet)
                if repair_fleet_order and repair_fleet_order.is_valid():
                    self.orders.append(repair_fleet_order)
            cur_fighter_capacity, max_fighter_capacity = FleetUtilsAI.get_fighter_capacity_of_fleet(fleet_id)
            if (
                fleet.fuel < fleet.maxFuel
                or cur_fighter_capacity < max_fighter_capacity
                and get_fleet_position(self.fleet.id) not in fleet_supplyable_system_ids
            ):
                resupply_fleet_order = MoveUtilsAI.get_resupply_fleet_order(self.fleet)
                if resupply_fleet_order.is_valid():
                    self.orders.append(resupply_fleet_order)
            return  # no targets

        if self.target:
            # for some targets fleet has to visit systems and therefore fleet visit them

            system_to_visit = (
                self.target.get_system()
                if not self.type == MissionType.PROTECT_REGION
                else TargetSystem(self._get_target_for_protection_mission())
            )
            if not system_to_visit:
                return
            orders_to_visit_systems = MoveUtilsAI.create_move_orders_to_system(self.fleet, system_to_visit)
            # TODO: if fleet doesn't have enough fuel to get to final target, consider resetting Mission
            for fleet_order in orders_to_visit_systems:
                self.orders.append(fleet_order)

            # also generate appropriate final orders
            fleet_order = self._get_fleet_order_from_target(
                self.type, self.target if not self.type == MissionType.PROTECT_REGION else system_to_visit
            )
            self.orders.append(fleet_order)

    def _need_repair(self, repair_limit: float = 0.70) -> bool:
        """Check if fleet needs to be repaired.

        If the fleet is already at a system where it can be repaired, stay there until fully repaired.
        Otherwise, repair if fleet health is below specified *repair_limit*.
        For military fleets, there is a special evaluation called, cf. *MilitaryAI.avail_mil_needing_repair()*

        :param repair_limit: percentage of health below which the fleet is sent to repair
        :return: True if fleet needs repair
        """
        # TODO: More complex evaluation if fleet needs repair (consider self-repair, distance, threat, mission...)
        fleet_id = self.fleet.id
        # if we are already at a system where we can repair, make sure we use it...
        system = self.fleet.get_system()
        # TODO starlane obstruction is not considered in the next call
        nearest_dock = MoveUtilsAI.get_best_drydock_system_id(system.id, fleet_id)
        if nearest_dock == system.id:
            repair_limit = 0.99
        # if combat fleet, use military repair check
        if get_aistate().get_fleet_role(fleet_id) in COMBAT_MISSION_TYPES:
            return (
                fleet_id
                in MilitaryAI.avail_mil_needing_repair(
                    [fleet_id], on_mission=bool(self.orders), repair_limit=repair_limit
                )[0]
            )
        # TODO: Allow to split fleet to send only damaged ships to repair
        ships_cur_health, ships_max_health = FleetUtilsAI.get_current_and_max_structure(fleet_id)
        return ships_cur_health < repair_limit * ships_max_health

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.fleet == other.target

    def __hash__(self):
        return hash(self.fleet)

    def __str__(self):
        fleet = self.fleet.get_object()

        fleet_id = self.fleet.id
        return "%-25s [%-11s] ships: %2d; total rating: %4d; target: %s" % (
            fleet,
            "NONE" if self.type is None else self.type,
            (fleet and len(fleet.shipIDs)) or 0,
            CombatRatingsAI.get_fleet_rating(fleet_id),
            self.target or "no target",
        )

    def _get_target_for_protection_mission(self):
        """Get a target for a PROTECT_REGION mission.

        1) If primary target (system target of this mission) is under attack, move to primary target.
        2) If neighbors of primary target have local enemy forces weaker than this fleet, may move to attack
        3) If no neighboring fleets or strongest enemy force is too strong, move to defend primary target
        """
        # TODO: Also check fleet rating vs planets in decision making below not only vs fleets
        universe = fo.getUniverse()
        primary_objective = self.target.id
        # TODO: Rate against specific threats
        fleet_rating = CombatRatingsAI.get_fleet_rating(self.fleet.id)
        debug(
            "%s finding target for protection mission (primary target %s). Fleet Rating: %.1f",
            self.fleet,
            self.target,
            fleet_rating,
        )
        immediate_threat = MilitaryAI.get_system_local_threat(primary_objective)
        if immediate_threat:
            debug("  Immediate threat! Moving to primary mission target")
            return primary_objective
        else:
            debug("  No immediate threats.")
            # Try to eliminate neighbouring fleets
            neighbors = get_neighbors(primary_objective)
            threat_list = sorted(map(lambda x: (MilitaryAI.get_system_local_threat(x), x), neighbors), reverse=True)

            if not threat_list:
                debug("  No neighbors (?!). Moving to primary mission target")
                return primary_objective
            else:
                debug("  Neighboring threats:")
                for threat, sys_id in threat_list:
                    debug("    %s - %.1f", TargetSystem(sys_id), threat)
            top_threat, candidate_system = threat_list[0]
            if not top_threat:
                # TODO: Move into second ring but needs more careful evaluation
                # For now, consider staying at the current location if enemy
                # owns a planet here which we can bombard.
                current_system_id = self.fleet.get_current_system_id()
                if current_system_id in neighbors:
                    system = universe.getSystem(current_system_id)
                    if assertion_fails(system is not None):
                        return primary_objective
                    empire_id = fo.empireID()
                    for planet_id in system.planetIDs:
                        planet = universe.getPlanet(planet_id)
                        if planet and not planet.ownedBy(empire_id) and not planet.unowned:
                            debug("Currently no neighboring threats. " "Staying for bombardment of planet %s", planet)
                            return current_system_id

                # TODO consider attacking neighboring, non-military fleets
                # - needs more careful evaluation against neighboring threats
                # empire_id = fo.empireID()
                # for sys_id in neighbors:
                #     system = universe.getSystem(sys_id)
                #     if assertion_fails(system is not None):
                #         continue
                #     local_fleets = system.fleetIDs
                #     for fleet_id in local_fleets:
                #         fleet = universe.getFleet(fleet_id)
                #         if not fleet or fleet.ownedBy(empire_id):
                #             continue
                #         return sys_id

                debug("No neighboring threats. Moving to primary mission target")
                return primary_objective

            # TODO rate against threat in target system
            # TODO only engage if can reach in 1 turn or leaves sufficient defense behind
            safety_factor = get_aistate().character.military_safety_factor()
            if fleet_rating < safety_factor * top_threat:
                debug("  Neighboring threat is too powerful. Moving to primary mission target")
                return primary_objective  # do not engage!

            debug("  Engaging neighboring threat: %s", TargetSystem(candidate_system))
            return candidate_system



from focs._effects import (
    AsteroidsType,
    ContainedBy,
    Contains,
    EffectsGroup,
    Focus,
    Happiness,
    InSystem,
    NamedReal,
    OwnedBy,
    Planet,
    Population,
    SetTargetIndustry,
    Source,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_AFTER_2ND_SCALING_PRIORITY

Tech(
    name="PRO_MICROGRAV_MAN",
    description="PRO_MICROGRAV_MAN_DESC",
    short_description="INDUSTRY_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=80 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["CON_ORBITAL_CON"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & ~Population(high=0)
            & Focus(type=["FOCUS_INDUSTRY"])
            & Happiness(low=NamedReal(name="PRO_MICROGRAV_MAN_MIN_STABILITY", value=10))
            & InSystem()
            & ContainedBy(Contains(Planet() & Planet(type=[AsteroidsType]) & OwnedBy(empire=Source.Owner))),
            priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
            effects=SetTargetIndustry(value=Value + NamedReal(name="PRO_MICROGRAV_MAN_TARGET_INDUSTRY_FLAT", value=2)),
        )
    ],
    graphic="icons/tech/microgravity_manufacturing.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from techs.ship_parts.fuel.fuel import PART_UPGRADE_MAXFUEL_EFFECTS

Tech(
    name="SHP_ANTIMATTER_TANK",
    description="SHP_ANTIMATTER_TANK_DESC",
    short_description="SHIP_FUEL_IMPROVE_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=270 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_FUEL_PART_TECHS"],
    prerequisites=["SHP_DEUTERIUM_TANK", "LRN_FORCE_FIELD"],
    effectsgroups=[PART_UPGRADE_MAXFUEL_EFFECTS("SHP_ANTIMATTER_TANK_EFFECT", 1.5)],
    graphic="icons/ship_parts/antimatter_tank.png",
)



import freeorion as fo

#################################
# Galaxy Generation Data Tables #
#################################

# For non-coders:
# If you aren't adding a new table, you only need to change the numbers.
# The order of rows doesn't matter, but the order of columns does (they should
# always be the same order as in universe/Enums.h).

# How are these numbers used?
# For each star type (including no star / deep space), a random number between
# 0 and 100 is chosen and all applicable bonuses are applied. Whichever star
# type gets the highest random number + fixed bonus is chosen for that system.
#
# The same system also gets used for planets, once all star types have been chosen.

# A simple one-dimensional table; in every roll in every galaxy, the star type
# before the colon gets the specified bonus.
BASE_STAR_TYPE_DIST = {
    fo.starType.blue: 15,
    fo.starType.white: 20,
    fo.starType.yellow: 25,
    fo.starType.orange: 20,
    fo.starType.red: 15,
    fo.starType.neutron: 0,
    fo.starType.blackHole: -5,
    fo.starType.noStar: 10,
}

# More complicated: These bonuses depend not only on the star type, but also
# on the value chosen on the galaxy setup screen for Age. If you look at it as
# a table, the entry in the "Young" row and "blue" column is the bonus for blue
# stars if you chose a Young galaxy.
UNIVERSE_AGE_MOD_TO_STAR_TYPE_DIST = {
    #                                blue  white  yellow  orange  red  neutron  blackHole  noStar
    #                        Young:
    fo.galaxySetupOptionGeneric.low: (10, 20, 0, 0, 0, 0, 0, 60),
    #                        Mature:
    fo.galaxySetupOptionGeneric.medium: (-5, 0, 10, 20, 0, 10, 10, 40),
    #                        Ancient:
    fo.galaxySetupOptionGeneric.high: (-20, -10, 0, 10, 20, 20, 20, 20),
}

# These following tables are as above, but for planet sizes rather than star
# types. The star types are now the givens, and the planet sizes are being
# rolled for.
#
# Note that for these purposes, Asteroids and Gas Giants are a size of planet.
#
# This checks what you chose for galaxy planet density:
DENSITY_MOD_TO_PLANET_SIZE_DIST = {
    #                                none  tiny  small  medium  large  huge  asteroids  gas giant
    fo.galaxySetupOptionGeneric.low: (85, 0, 0, 0, -5, -10, 0, 0),
    fo.galaxySetupOptionGeneric.medium: (70, 0, 0, 0, -5, -10, 0, 0),
    fo.galaxySetupOptionGeneric.high: (50, 0, 0, 0, -5, -10, 0, 0),
}

# Given the star type that was already chosen, what will the bonus be for this
# size of planet?
# Note: A positive value for none, will mean a chance of empty systems
# of that type in game.  All star types with a zero or less chance for
# none will be forced to have at least one satellite.
STAR_TYPE_MOD_TO_PLANET_SIZE_DIST = {
    #                           none  tiny  small  medium  large  huge  asteroids  gas giant
    fo.starType.blue: (0, 0, 0, 0, -5, -10, 0, 0),
    fo.starType.white: (0, 0, 0, 0, -5, -10, 0, 0),
    fo.starType.yellow: (0, 0, 0, 0, -5, -10, 0, 0),
    fo.starType.orange: (0, 0, 0, 0, -5, -10, 0, 0),
    fo.starType.red: (0, 0, 0, 0, -5, -10, 0, 0),
    fo.starType.neutron: (5, 0, 0, 0, -5, -10, 20, 0),
    fo.starType.blackHole: (30, 0, 0, -5, -10, -15, 10, -5),
    fo.starType.noStar: (80, 0, 0, 0, -10, -15, 0, 10),
}

# In this one, the given is the orbit number of the planet. So the first row
# corresponds to the planet closest to the star, the second is the next one,
# and so on.
ORBIT_MOD_TO_PLANET_SIZE_DIST = (
    #    none  tiny  small  medium  large  huge  asteroids  gas giant
    (10, 0, 20, 10, 0, 0, 0, -80),
    (5, 0, 15, 20, 5, 0, 0, -70),
    (0, 0, 10, 30, 10, 5, 0, -60),
    (0, 0, 10, 30, 10, 5, 10, -30),
    (0, 0, 10, 20, 15, 10, 25, 0),
    (0, 0, 5, 5, 20, 15, 10, 30),
    (0, 0, 5, 5, 15, 10, 5, 30),
    (0, 0, 10, 5, 10, 5, 5, 30),
    (5, 0, 10, 5, 5, 5, 5, 30),
    (10, 0, 10, 0, 0, 0, 5, 10),
)

# Bonus to the planet size based on the shape of the galaxy.
GALAXY_SHAPE_MOD_TO_PLANET_SIZE_DIST = {
    #                                none  tiny  small  medium  large  huge  asteroids  gas giant
    fo.galaxyShape.spiral2: (0, 0, 0, 0, 0, 0, 0, 0),
    fo.galaxyShape.spiral3: (0, 0, 0, 0, 0, 0, 0, 0),
    fo.galaxyShape.spiral4: (0, 0, 0, 0, 0, 0, 0, 0),
    fo.galaxyShape.cluster: (0, 0, 0, 0, 0, 0, 0, 0),
    fo.galaxyShape.elliptical: (0, 0, 0, 0, 0, 0, 0, 0),
    fo.galaxyShape.disc: (0, 0, 0, 0, 0, 0, 0, 0),
    fo.galaxyShape.box: (0, 0, 0, 0, 0, 0, 0, 0),
    fo.galaxyShape.irregular: (0, 0, 0, 0, 0, 0, 0, 0),
    fo.galaxyShape.ring: (0, 0, 0, 0, 0, 0, 0, 0),
}

# Whenever any special might be placed, the following probability is the chance
# that it actually will be (after other factors are considered).
SPECIALS_FREQUENCY = {
    fo.galaxySetupOptionGeneric.none: 0.0,
    fo.galaxySetupOptionGeneric.low: 0.08,
    fo.galaxySetupOptionGeneric.medium: 0.2,
    fo.galaxySetupOptionGeneric.high: 0.5,
}

# Whenever a monster might be placed, the following probability is the chance
# that it actually will be (after other factors are considered).
MONSTER_FREQUENCY = {
    fo.galaxySetupOptionMonsterFreq.none: 0.0,
    fo.galaxySetupOptionMonsterFreq.extremelyLow: 0.002,
    fo.galaxySetupOptionMonsterFreq.veryLow: 0.01,
    fo.galaxySetupOptionMonsterFreq.low: 0.03,
    fo.galaxySetupOptionMonsterFreq.medium: 0.08,
    fo.galaxySetupOptionMonsterFreq.high: 0.2,
    fo.galaxySetupOptionMonsterFreq.veryHigh: 0.5,
    fo.galaxySetupOptionMonsterFreq.extremelyHigh: 0.999,
}

# Whenever natives might be placed on a planet, the following probability is
# the chance that they actually will be (after other factors are considered).
NATIVE_FREQUENCY = {
    fo.galaxySetupOptionGeneric.none: 0.0,
    fo.galaxySetupOptionGeneric.low: 0.05,
    fo.galaxySetupOptionGeneric.medium: 0.125,
    fo.galaxySetupOptionGeneric.high: 0.3,
}

# This is the maximum size of the shortest path via starlane that the generator
# can create between two "adjacent" systems. So when the number is smaller, the
# map will end up with more starlanes to create enough connections between adjacent
# systems.
MAX_JUMPS_BETWEEN_SYSTEMS = {
    fo.galaxySetupOptionGeneric.low: 8,
    fo.galaxySetupOptionGeneric.medium: 3,
    fo.galaxySetupOptionGeneric.high: 1,
}

# The maximum length of any starlane the galaxy generator can create (in uu).
# The maximum length enforcment only works if there is at least 2*MIN_SYSTEM_SEPARATION
# between the nearest stars in two groups, in order to place extra star(s) in between.
# Consequently, MAX_STARLANE_LENGTH >= 2 * MIN_SYSTEM_SEPARATION
MAX_STARLANE_LENGTH = 120

# This is the minimum distance between two star systems (in uu).
MIN_SYSTEM_SEPARATION = 35.0



from focs._effects import (
    Capital,
    EffectsGroup,
    GameRule,
    HasSpecies,
    LocalCandidate,
    MaxOf,
    MinOf,
    Planet,
    Population,
    SetSpeciesOpinion,
    SetSpeciesTargetOpinion,
    SpeciesEmpireTargetOpinion,
    SpeciesShipsDestroyed,
    SpeciesShipsLost,
    StatisticCount,
    Target,
    ThisSpecies,
    Value,
)


def FIXED_OPINION_EFFECTS(name: str, value: float):
    return EffectsGroup(
        scope=Capital,
        stackinggroup="%s_FIXED_OPINION_OF_EMPIRE" % name,
        effects=[
            SetSpeciesTargetOpinion(species=ThisSpecies, empire=Target.Owner, opinion=value),
            SetSpeciesOpinion(species=ThisSpecies, empire=Target.Owner, opinion=value),
        ],
    )


def COMMON_OPINION_EFFECTS(name: str):
    return EffectsGroup(
        scope=Capital,
        stackinggroup="%s_TOWARDS_TARGET_OPINION_OF_EMPIRE" % name,
        effects=[
            SetSpeciesTargetOpinion(  # baseline opinion
                species=ThisSpecies,
                empire=Target.Owner,
                opinion=Value
                + GameRule(type=float, name="RULE_BASELINE_SPECIES_EMPIRE_OPINION")
                + (
                    GameRule(type=float, name="RULE_INVASION_OPINION_PENALTY_SCALING")
                    * StatisticCount(
                        float,
                        condition=Planet()
                        & HasSpecies(name=[ThisSpecies])
                        & (LocalCandidate.LastInvadedByEmpire == Target.Owner),
                    )
                    ** 0.5
                )
                + (
                    GameRule(type=float, name="RULE_SHIPS_LOST_DESTROYED_PENALTY_SCALING")
                    * (
                        SpeciesShipsDestroyed(empire=Target.Owner, name=ThisSpecies)
                        + SpeciesShipsLost(empire=Target.Owner, name=ThisSpecies)
                    )
                    ** 0.5
                )
                + (
                    GameRule(type=float, name="RULE_COLONIES_FOUNDED_BONUS_SCALING")
                    * StatisticCount(
                        float,
                        condition=Planet()
                        & HasSpecies(name=[ThisSpecies])
                        & Population(low=0.001)
                        & (LocalCandidate.LastColonizedByEmpire == Target.Owner),
                    )
                ),
            ),
            SetSpeciesOpinion(
                species=ThisSpecies,
                empire=Target.Owner,
                opinion=Value
                + MinOf(
                    float,
                    1.0,  # increase by 1 per turn up towards target
                    MaxOf(
                        float,
                        -1.0,  # decrease by 1 per turn down towards target
                        SpeciesEmpireTargetOpinion(species=ThisSpecies, empire=Target.Owner) - Value,
                    ),
                ),
            ),
        ],
    )



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TOXIC_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import BAD_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import GREAT_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS, NO_OFFENSE_TROOPS
from species.species_macros.weapons import BAD_WEAPONS

Species(
    name="SP_DERTHREAN",
    description="SP_DERTHREAN_DESC",
    gameplay_description="SP_DERTHREAN_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "TELEPATHIC",
        "GREAT_INFLUENCE",
        "NO_OFFENSE_TROOPS",
        "BAD_WEAPONS",
        "BAD_HAPPINESS",
        "AVERAGE_SUPPLY",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "PEDIA_TELEPATHIC_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "MIMETIC_ALLOY_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "GAIA_SPECIAL",
        "PHILOSOPHER_SPECIAL",
        "RESONANT_MOON_SPECIAL",
        "COMPUTRONIUM_SPECIAL",
        "PLC_DIVINE_AUTHORITY",
        "PLC_DIVERSITY",
        "PLC_CONFEDERATION",
        "PLC_MODERATION",
        "PLC_INDOCTRINATION",
        "PLC_INDUSTRIALISM",
    ],
    dislikes=[
        "BLD_SHIPYARD_ORG_CELL_GRO_CHAMB",
        "BLD_SHIPYARD_ENRG_COMP",
        "BLD_NEUTRONIUM_SYNTH",
        "BLD_NEUTRONIUM_FORGE",
        "BLD_NEUTRONIUM_EXTRACTOR",
        "MINERALS_SPECIAL",
        "CRYSTALS_SPECIAL",
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "FORTRESS_SPECIAL",
        "PLC_CONFORMANCE",
        "PLC_CENTRALIZATION",
        "PLC_TERROR_SUPPRESSION",
        "PLC_NO_GROWTH",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *AVERAGE_RESEARCH,
        *GREAT_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_DERTHREAN"),
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        *NO_OFFENSE_TROOPS,
        *BAD_WEAPONS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TOXIC_STANDARD_EP,
    graphic="icons/species/derthrean.png",
)



"""
Character represents the personalization of the AI.

Character is composed of orthogonal elements called Trait.  Trait
elements are orthogonal, which means that they do not interact and can be
freely mixed and matched to create a Character.  Orthogonality means
that they can be tested independently and combined with simple
combiners.

Character creates big changes in the trait of the AI that are visible
to the player as personality.  At key/tap points the AI checks the character
for permission 'may_<do something>' or preference of a series of
alternative actions 'prefer_<somethings>'.  Preference could be expanded
to look at all of the options and remove those that the trait
forbids.

Permission type taps are named may_<do something>(information needed to decide).
The trait examines the information and returns true if that action is permitted.

Preference type taps are named prefer_<somethings>(alternatives, extra info).
Alternatives is a list of all possible actions.  The trait examines all
the alternatives and returns a possibly empty list of the permissible
actions.

Character should be invoked by other modules at key trait tap points to
provide direction not optimization.  For example which empire/species do
I love/hate enough to attack.

The gated traits should be big user discernable changes, not small
coefficient optimizations better handled with a local optimization
routine. If there are setup options to the trait they should fall
into 2 (on/off) or 3 settings.  For example,
deceitful/typical/trustworthy.  The idea is not to create deep, subtle,
nuanced personalities, but writ large Shakespearean characters that draw the
player into the narrative structure of the particular game that they are playing.

Each AI will get the mandatory trait (probably Difficulty/Challenge)
and a selection of the optional traits.

There is no need for per species trait.  Each playable species can be
assigned some additional mandatory trait(s).  For example the Trith and
a mandatory Genocidal trait.  Perhaps add a probabilty distribution of
trait components to the FOCS description of a playable species.
"""
# Some ideas for future trait modules are:
# TODO: challenge/difficulty -- to replace the difficulty related portions
#                                of aggression.
#
# TODO: research bent -- models an obsession/focus/repulsion by a particular
# area of research. i.e. Shields are for grues.
#
# TODO: nemesis/single-minded -- model a focus on a single opponent, perhaps
# the first one to attack the AI.  This plays well with CharacterStrings
# which can be used to inform all players who care that "I will hunt the
# Dominion to the ends of the galaxy."  Humans and eventually AIs could
# use this to modify their risk assessment of this AI.
#
# TODO: deceitful/trustworthy -- Only useful after treaties are implemented to
# determine if this AI will abide by the treaty
#
# TODO: friendly and loyal -- Even without treaties this AI will assist their
# friends.
#
# TODO: vengeful -- An eye for eye.  AI generally counter attacks until things
# are made right or even.
#
# TODO: risk averse/gambler -- How much overkill does the AI require?  How risky
# a research strategy will it attempt?  How many redundant buildings does
# the AI build?
#
# TODO: warlike/peaceful -- Will the AI prefer colonization to war?
#
# TODO: cooperative/loner -- Once cooperative treaties (joint attack, sharing
# system) are implemented, will the AI use them?
#
# TODO: taciturn/chatty -- How often does the AI chat?
#
# TODO: genocidal -- Will the AI research and use Concentration Camps/bombard
# weapons/planet and star destroying weapons?  Is the AI horrified by the
# use of such weapons?  Will the AI band together with other like minded
# empires to protect the galaxy?
#

import abc
import freeOrionAIInterface as fo
import math
import random
from collections import Counter
from logging import debug, warning


class Trait(metaclass=abc.ABCMeta):
    """An abstract class representing a type of trait of the AI.

    Traits give the AI personality along some dimension.
    Traits do not help the AI make optimal decisions, they determine whether
    certain actions are permissible or preferable.

    Traits are combined to form a single Character.

    Traits have taps which the AI calls to determine it trait with
    respect to a single action.  There are two types of taps: permission
    taps which permit/forbid an action and preference taps which indicate
    which of several alternatives are permissible.

    Permission type taps are named may_<do something>(information needed to decide).
    The trait examines the information and returns true if that action is
    permitted.

    Preference type taps are named prefer_<somethings>(alternatives, extra info).
    Alternatives is a list of all possible actions.  The trait examines all of
    the alternatives and returns a possibly empty list of the permissible actions.

    Any given Trait class should not implement all the taps, only those
    it needs to override to cause the relevant trait.
    """

    def __repr__(self):
        return "Trait"

    # @abc.abstractproperty
    @property
    def key(self):  # pylint: disable=no-self-use,unused-argument
        """Return a key to be used as an index into look up tables.

        For example, a string table of 6 diplomatic responses bases on aggression might look like:

        table_x = {fo.aggression.beginner: "DIPLO_X_BEGINNER",
                   fo.aggression.turtle: "DIPLO_X_TURTLE",
                   fo.aggression.cautious: "DIPLO_X_CAUTIOUS",
                   fo.aggression.typical: "DIPLO_X_TYPICAL",
                   fo.aggression.aggressive: "DIPLO_X_AGGRESSIVE",
                   fo.aggression.maniacal: "DIPLO_X_MANIACAL"}

        Using the key to fetch a single string from that table_x looks like:

        used_string_x = table_x[character.get_trait(Character.Trait.Aggression).key]

        See character_strings_module.py for the details of this actual example.

        """
        return None

    def may_explore_system(self, monster_threat):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to explore system with the given monster threat."""
        return True

    def may_surge_industry(self, totalPP, totalRP):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to surge industry as used in PriorityAI.py"""
        return True

    def may_maximize_research(self):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to maximize research."""
        return True

    def preferred_research_cutoff(self, alternatives):  # pylint: disable=no-self-use,unused-argument
        """Return preferred research cutoff from the list of alternatives."""
        # TODO Remove this tap it is one of the empire id dependent taps. See EmpireIDTrait.
        return None

    def max_number_colonies(self):  # pylint: disable=no-self-use,unused-argument
        """Return maximum allowed number of colonies"""
        return 1000000

    def may_invade(self):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to invade."""
        return True

    def may_invade_with_bases(self):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to invade with bases."""
        return True

    def invasion_priority_scaling(self):  # pylint: disable=no-self-use,unused-argument
        return 1.0

    def military_priority_scaling(self):  # pylint: disable=no-self-use,unused-argument
        return 1.0

    def preferred_colonization_portion(self, alternatives):  # pylint: disable=no-self-use,unused-argument
        """Select from the fractions in alternatives the fraction of PP to be spend on colonization."""
        # TODO Remove this tap it is one of the empire id dependent taps. See EmpireIDTrait.
        return None

    def preferred_outpost_portion(self, alternatives):  # pylint: disable=no-self-use,unused-argument
        """Select from the fractions in alternatives the fraction of PP to be spend on outposts."""
        # TODO Remove this tap it is one of the empire id dependent taps. See EmpireIDTrait.
        return None

    def preferred_building_ratio(self, alternatives):  # pylint: disable=no-self-use,unused-argument
        """Select a fraction less than 1 from alternatives as the maximum ratio of PP for buildings"""
        # TODO Remove this tap it is one of the empire id dependent taps. See EmpireIDTrait.
        return None

    def preferred_discount_multiplier(self, alternatives):  # pylint: disable=no-self-use,unused-argument
        """Select a discount multiplier from the alternatives provided for
        use in evaluate planet in Colonisation.py to scale pilot rating and
        a long list of technologies.
        """
        # TODO Remove this tap it is one of the empire id dependent taps. See EmpireIDTrait.
        return None

    def max_defense_portion(self):  # pylint: disable=no-self-use,unused-argument
        """Return maximum fraction of production for defense"""
        return 1.0

    def check_orbital_production(self):  # pylint: disable=no-self-use,unused-argument
        """Return true if orbital defense production should be checked this turn in ProductionAI.py"""
        return False

    def target_number_of_orbitals(self):  # pylint: disable=no-self-use,unused-argument
        """Return target number of orbitals defenses to be built."""
        return 0

    def may_build_building(self, building):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to build ''building''"""
        return True

    def may_produce_troops(self):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to produce troop ships"""
        return True

    def military_safety_factor(self):  # pylint: disable=no-self-use,unused-argument
        """Return military safety factor, the ratio of
        (enemy strength) / (own strength) that the AI considers acceptable risk.
        """
        return 0.0

    def may_research_heavily(self):  # pylint: disable=no-self-use,unused-argument
        """Return true if allowed to target research/industry > 1.5"""
        return True

    def may_use_growth_focus(self):
        """Return True if permitted to use growth focus."""
        return True

    def may_travel_beyond_supply(self, distance):  # pylint: disable=no-self-use,unused-argument
        """Return True if able to travel distance hops beyond empire supply"""
        # TODO Remove this tap it is one of the empire id dependent taps. See EmpireIDTrait.
        return True

    def get_research_index(self):  # pylint: disable=no-self-use,unused-argument
        """Deprecated.  Only used with old style research."""
        # TODO Remove this tap when old style research is removed.
        return None

    def may_research_xeno_genetics_variances(self):  # pylint: disable=no-self-use,unused-argument
        """Return True if AI if allowed to research xeno genetics research 'Dep.GRO_XENO_GENETICS'."""
        # TODO remove this as overly specific
        return True

    def prefer_research_defensive(self):  # pylint: disable=no-self-use,unused-argument
        """Return True if should prefer defensive tech research"""
        return True

    def prefer_research_low_aggression(self):  # pylint: disable=no-self-use,unused-argument
        """Return True if should prefer less aggressive tech research"""
        return True

    def may_research_tech(self, tech):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to research ''tech''. This is only called by the new research algorithm."""
        return True

    def may_research_tech_classic(self, tech):  # pylint: disable=no-self-use,unused-argument
        """Return True if permitted to research ''tech''.  This is called in the classic research algorithm."""
        # TODO remove this tap when the classic research algorithm is removed.
        return True

    def attitude_to_empire(self, other_empire_id, diplomatic_logs):  # pylint: disable=no-self-use,unused-argument
        """Return a value from [-10 .. 10] representing attitude towards other empire."""
        return None

    def warship_adjusted_production_cost_exponent(self):  # pylint: disable=no-self-use,unused-argument
        """Return an exponent to scale the production cost of a warship in ShipDesignAI."""
        return None

    def secondary_valuation_factor_for_invasion_targets(self):  # pylint: disable=no-self-use,unused-argument
        """Return a value in range [0.0 : 1.0], used in colonization scoring calculations where subscores for a primary
        planet depend on traits of secondary planets, for what portion of the subscore should be assigned if the secondary
        planet would need to be acquired via invasion"""
        return None


class Aggression(Trait):
    """A trait that models level of difficulty and aggression."""

    # The initial implementation was to pull aggression dependent code from
    # the main body of code into this trait Aggression.  These are merely
    # refactoring of existing code and is not an example of ideal
    # implementation.

    # Aggression implements two different traits as one: difficulty and
    # aggression.  It creates too many decision points not clearly related to
    # either concept.  Aggression should be broken into two traits:
    # difficulty/challenge and aggression.

    # TODO break this class into two traits: level of difficulty and aggression

    def __init__(self, aggression):
        self.aggression = aggression

    @property
    def key(self):
        return self.aggression

    def may_explore_system(self, monster_threat):
        return monster_threat < 2000 * self.aggression

    def may_surge_industry(self, total_pp, total_rp):
        return (self.aggression > fo.aggression.cautious) and ((total_pp + 1.6 * total_rp) < (50 * self.aggression))

    def may_maximize_research(self):
        return self.aggression >= fo.aggression.maniacal

    def max_number_colonies(self):
        # significant growth barrier for low aggression, negligible for high aggression
        # TODO: consider further changes, including a dependency on galaxy size and planet density
        return 2 + ((0.5 + 1.4 * self.aggression) ** 2) * fo.currentTurn() / 50.0

    def may_invade(self):
        return self.aggression > fo.aggression.turtle and not (
            self.aggression == fo.aggression.beginner and fo.currentTurn() < 150
        )

    def may_invade_with_bases(self):
        return self.aggression > fo.aggression.typical

    def invasion_priority_scaling(self):
        return 0.5 if self.aggression == fo.aggression.beginner else 1.0

    def military_priority_scaling(self):
        return (
            1.0
            if self.aggression > fo.aggression.typical
            else ((1.0 + self.aggression) / (1.0 + fo.aggression.typical))
        )

    def max_defense_portion(self):
        return [0.7, 0.4, 0.3, 0.2, 0.1, 0.0][self.aggression]

    def check_orbital_production(self):
        aggression_index = max(1, self.aggression)
        return ((fo.currentTurn() % aggression_index) == 0) and self.aggression < fo.aggression.maniacal

    def target_number_of_orbitals(self):
        aggression_index = max(1, self.aggression)
        return min(
            int(((fo.currentTurn() + 4) / (8.0 * aggression_index**1.5)) ** 0.8),
            fo.aggression.maniacal - aggression_index,
        )

    BUILDING_TABLE_STATIC = {
        "BLD_SHIPYARD_AST": fo.aggression.beginner,
        "BLD_GAS_GIANT_GEN": fo.aggression.beginner,
        "BLD_SOL_ORB_GEN": fo.aggression.turtle,
        "BLD_ART_BLACK_HOLE": fo.aggression.typical,
        "BLD_BLACK_HOLE_POW_GEN": fo.aggression.cautious,
        "BLD_CONC_CAMP": fo.aggression.typical,
        "BLD_SHIPYARD_CON_NANOROBO": fo.aggression.aggressive,
        "BLD_SHIPYARD_CON_GEOINT": fo.aggression.aggressive,
        # TODO determine which duplicate of BLD_SHIPYARD_AST is correct
        # "BLD_SHIPYARD_AST": fo.aggression.typical,
        "BLD_SHIPYARD_AST_REF": fo.aggression.maniacal,
        "BLD_SHIPYARD_ORG_CELL_GRO_CHAMB": fo.aggression.aggressive,
        "BLD_SHIPYARD_ORG_XENO_FAC": fo.aggression.aggressive,
        "BLD_SHIPYARD_ENRG_COMP": fo.aggression.aggressive,
        "BLD_SHIPYARD_ENRG_SOLAR": fo.aggression.maniacal,
        "BLD_NEUTRONIUM_FORGE": fo.aggression.cautious,
    }

    @property
    def building_table(self):
        return type(self).BUILDING_TABLE_STATIC

    def may_build_building(self, building):
        return self.aggression > self.building_table.get(building, fo.aggression.beginner)

    def may_produce_troops(self):
        # TODO check if this is consitent with may invade
        return self.aggression >= fo.aggression.typical

    def military_safety_factor(self):
        return [4.0, 3.0, 2.0, 1.5, 1.2, 1.0][self.aggression]

    def may_research_heavily(self):
        return self.aggression > fo.aggression.cautious

    def may_use_growth_focus(self):
        # For now, allow using growth focus for all aggression settings
        # but leaving this here for easier modification.
        return self.aggression >= fo.aggression.beginner

    def may_research_xeno_genetics_variances(self):
        return self.aggression >= fo.aggression.cautious

    def prefer_research_defensive(self):
        return self.aggression <= fo.aggression.cautious

    def prefer_research_low_aggression(self):
        return self.aggression < fo.aggression.typical

    tech_lower_threshold_static = {"LRN_TRANSCEND": fo.aggression.aggressive}

    def may_research_tech(self, tech):
        return type(self).tech_lower_threshold_static.get(tech, fo.aggression.beginner) <= self.aggression

    TECH_UPPER_THRESHOLD_CLASSIC_STATIC = {
        "DEF_DEFENSE_NET_1": fo.aggression.cautious,
        "DEF_GARRISON_1": fo.aggression.cautious,
        "GRO_XENO_GENETICS": fo.aggression.cautious,
        "GRO_GENETIC_ENG": fo.aggression.cautious,
    }

    TECH_LOWER_THRESHOLD_CLASSIC_STATIC = {
        "SHP_DEFLECTOR_SHIELD": fo.aggression.aggressive,
        "CON_ARCH_PSYCH": fo.aggression.aggressive,
        "CON_CONC_CAMP": fo.aggression.aggressive,
        "LRN_XENOARCH": fo.aggression.typical,
        "LRN_PHYS_BRAIN": fo.aggression.typical,
        "LRN_TRANSLING_THT": fo.aggression.typical,
        "LRN_PSIONICS": fo.aggression.typical,
        "LRN_DISTRIB_THOUGHT": fo.aggression.typical,
        "LRN_QUANT_NET": fo.aggression.typical,
        "PRO_SINGULAR_GEN": fo.aggression.typical,
        "LRN_TRANSCEND": fo.aggression.typical,
    }

    def may_research_tech_classic(self, tech):
        return (
            type(self).TECH_LOWER_THRESHOLD_CLASSIC_STATIC.get(tech, fo.aggression.beginner)
            <= self.aggression
            <= type(self).TECH_UPPER_THRESHOLD_CLASSIC_STATIC.get(tech, fo.aggression.maniacal)
        )

    def attitude_to_empire(self, other_empire_id, diplomatic_logs):
        # TODO: In other traits consider proximity, competitive
        # needs, relations with other empires, past history with this
        # empire, etc.
        # in the meantime, somewhat random
        # TODO: Move the diplomatic log portion of this trait back
        # into diplomacy where it belongs.
        if self.aggression == fo.aggression.maniacal:
            return -9
        if self.aggression == fo.aggression.beginner:
            return 9
        log_index = (other_empire_id, fo.empireID())
        num_alliance_requests = len(diplomatic_logs.get("alliance_requests", {}).get(log_index, []))
        num_peace_requests = len(diplomatic_logs.get("peace_requests", {}).get(log_index, []))
        num_war_declarations = len(diplomatic_logs.get("war_declarations", {}).get(log_index, []))
        # Too many requests for peace irritate the AI, as do any war declarations
        irritation = (
            self.aggression
            * (2.0 + num_alliance_requests / 5.0 + num_peace_requests / 10.0 + 2.0 * num_war_declarations)
            + 0.5
        )
        attitude = 10 * random.random() - irritation
        return min(10, max(-10, attitude))

    def warship_adjusted_production_cost_exponent(self):  # pylint: disable=no-self-use,unused-argument
        # as military ships are grouped up in fleets, their power rating scales quadratic in numbers.
        # To account for this, we need to maximize rating/cost_squared not rating/cost as usual.
        if self.aggression == fo.aggression.maniacal:
            exponent = 2.0
        elif self.aggression == fo.aggression.aggressive:
            exponent = 1.5
        else:
            exponent = 1.0
        return exponent

    def secondary_valuation_factor_for_invasion_targets(self):  # pylint: disable=no-self-use,unused-argument
        """Return a value in range [0.0 : 1.0], used in colonization scoring calculations where subscores for a primary
        planet depend on traits of secondary planets, for what portion of the subscore should be assigned if the secondary
        planet would need to be acquired via invasion"""
        if self.aggression == fo.aggression.maniacal:
            factor = 0.8
        elif self.aggression == fo.aggression.aggressive:
            factor = 0.4
        elif self.aggression == fo.aggression.typical:
            factor = 0.2
        else:
            factor = 0.0
        return factor


class EmpireIDTrait(Trait):
    """A trait that models empire id influence.
    Mostly some modulo 2 effects."""

    # The initial implementation was to pull empire.id modulo 2 or 3 optional
    # portions of the code into this trait EmpireIDTrait.  This was
    # merely a refactoring of existing code and is not an example of ideal
    # implementation.

    # EmpireID selects based on empire.id with is an implementation detail and
    # not a "visible" game mechanic.  EmpireID is also used to select between
    # coefficient options, which look like programmer experiments.  EmpireID
    # should be removed as a trait.

    # TODO: Remove EmpireIDTrait.
    # Empire id is a game mechanic.  It is not something that a player
    # can describe, "Look the 'Continuum' is behaving like a 1 modulo 2 character."

    def __init__(self, empire_id, aggression):
        debug("EmpireIDTrait initialized.")
        self.id = empire_id
        self.aggression = aggression  # TODO remove when old research style get_research_index is removed

    @property
    def key(self):
        return self.id % 2

    def _modulo_two_choice(self, alternatives):
        if not alternatives:
            return None
        if len(alternatives) == 1:
            return alternatives[0]
        return alternatives[self.id % 2]

    def preferred_research_cutoff(self, alternatives):
        return self._modulo_two_choice(alternatives)

    def preferred_colonization_portion(self, alternatives):
        return self._modulo_two_choice(alternatives)

    def preferred_outpost_portion(self, alternatives):
        return self._modulo_two_choice(alternatives)

    def preferred_building_ratio(self, alternatives):
        if not alternatives:
            return None
        return alternatives[self.id % 3] if len(alternatives) >= 3 else alternatives[self.id % len(alternatives)]

    def preferred_discount_multiplier(self, alternatives):
        return self._modulo_two_choice(alternatives)

    def may_travel_beyond_supply(self, distance):
        return (
            distance < 2
            or distance <= 2
            and self.aggression >= fo.aggression.typical
            or self.aggression >= fo.aggression.aggressive
        )

    # TODO remove this function as soon as old style research is gone
    # It defeats the orthogonality goal of character components for little gain
    def get_research_index(self):
        research_index = self.id % 2
        if self.aggression >= fo.aggression.aggressive:
            research_index = 2 + (self.id % 3)  # so indices [2,3,4]
        elif self.aggression >= fo.aggression.typical:
            research_index += 1
        return research_index


class Character(Trait):
    """
    A collection of traits.

    For each query that Trait supports a Character queries
    all of its own traits to determine if an action is permissible or prefered.
    """

    # See Note below: Algorithmic Completion of Character class.
    def __init__(self, traits):
        self.traits = traits
        if not all([isinstance(x, Trait) for x in traits]):
            raise TypeError("All traits must be sub-classes of Trait")

    def get_trait(self, type_of_trait):
        """Return the requested trait or None"""
        trait = [x for x in self.traits if isinstance(x, type_of_trait)]
        return trait[0] if trait else None


# Note: Algorithmic Completion of Character class.
# Complete the Character class by adding all of the combiners to combine the outputs of the
# individual traits.  Character tries to combine results in the way most limiting to the AI


def _make_single_function_combiner(funcnamei, f_combo):
    """Make a combiner that collects the results of funcname from each trait
    and applies f_combo to the results"""

    def func(self, *args, **kwargs):
        """Apply funcnamei to each trait and combine them with ''f_combo''"""
        return f_combo([getattr(x, funcnamei)(*args, **kwargs) for x in self.traits])

    return func


def _maxmin_not_none(f_combo):
    """Make a combiner that collects not None items and applies min or max"""

    def func(llin):
        ll = [x for x in llin if x is not None]
        if not ll:
            return 0
        return f_combo(ll)

    return func


# Create combiners for traits that all must be true
for funcname in [
    "may_explore_system",
    "may_surge_industry",
    "may_maximize_research",
    "may_invade",
    "may-invade_with_bases",
    "may_build_building",
    "may_produce_troops",
    "may_research_heavily",
    "may_use_growth_focus",
    "may_travel_beyond_supply",
    "may_research_xeno_genetics_variances",
    "prefer_research_defensive",
    "prefer_research_low_aggression",
    "may_research_tech",
    "may_research_tech_classic",
]:
    setattr(Character, funcname, _make_single_function_combiner(funcname, all))

# Create combiners for traits that take min result
for funcname in [
    "max_number_colonies",
    "invasion_priority_scaling",
    "military_priority_scaling",
    "max_defense_portion",
]:
    setattr(Character, funcname, _make_single_function_combiner(funcname, _maxmin_not_none(min)))

# Create combiners for traits that take max result
for funcname in ["target_number_of_orbitals", "military_safety_factor", "get_research_index"]:
    setattr(Character, funcname, _make_single_function_combiner(funcname, _maxmin_not_none(max)))

# Create combiners for traits that take any result
for funcname in ["check_orbital_production"]:
    setattr(Character, funcname, _make_single_function_combiner(funcname, any))


# Create combiners for traits that averages all not None results
def average_not_none(llin):
    ll = [x for x in llin if x is not None]
    if not ll:
        return 0
    return sum(ll) / float(len(ll))


for funcname in ["attitude_to_empire", "secondary_valuation_factor_for_invasion_targets"]:
    setattr(Character, funcname, _make_single_function_combiner(funcname, average_not_none))


# Create combiners for traits that use the geometic mean of all not None results
def geometric_mean_not_none(llin):
    ll_not_none = [x for x in llin if x is not None]
    ll = [x for x in ll_not_none if x > 0]
    if len(ll_not_none) != len(ll):
        warning("Calculating the geometric mean of %s contains negative numbers which will be ignored." % ll_not_none)
    if not ll:
        return 1
    return math.exp(sum(map(math.log, ll)) / float(len(ll)))


for funcname in ["warship_adjusted_production_cost_exponent"]:
    setattr(Character, funcname, _make_single_function_combiner(funcname, geometric_mean_not_none))


def _make_most_preferred_combiner(funcnamei):
    """Make a combiner that runs the preference function for each trait and
    returns the result most preferred by all the traits."""

    def _most_preferred(self, alternatives):
        """Applies funcnamei from each trait to the alternatives and return the most preferred."""
        prefs = [y for y in [getattr(x, funcnamei)(alternatives) for x in self.traits] if y is not None]
        if not prefs:
            return None
        if len(prefs) == 1:
            return prefs[0]
        return Counter.most_common(Counter(prefs), 1)[0][0]

    return _most_preferred


# Create combiners for traits deal with preference
for funcname in [
    "preferred_research_cutoff",
    "preferred_colonization_portion",
    "preferred_outpost_portion",
    "preferred_building_ratio",
    "preferred_discount_multiplier",
]:
    setattr(Character, funcname, _make_most_preferred_combiner(funcname))


def create_character(aggression=fo.aggression.maniacal, empire_id=0):
    """Create a character."""
    # TODO add the mandatory (Difficulty) and optional/random (everything
    # else) interface to Character creation.

    # Check the optionsDB for the trait bypass values and create
    # the character.
    NO_VALUE = -1
    bypassed_aggression = get_trait_bypass_value("aggression", int(aggression), NO_VALUE)
    bypassed_empire_id = get_trait_bypass_value("empire-id", empire_id, NO_VALUE)

    return Character([Aggression(bypassed_aggression), EmpireIDTrait(bypassed_empire_id, bypassed_aggression)])


def get_trait_bypass_value(name: str, default: int, sentinel: int) -> int:
    """Fetch a bypassed trait value or return the default from OptionsDB.

    In OptionsDB a section ai.config.trait can contain default trait
    values for all of the AIs or specific AIs which will override the
    default value passed into this function.

    If there is an XML element in config.xml/persistent_config.xml
    ai.trait.<name of trait here>.force.enabled
    with a non zero value

    ,then the value of ai.trait.<name of trait here>.ai_<AI ID number here>

    will be checked.  If it is not the sentinel value (typically -1) the it
    will be returned as the trait's value.

    Otherwise the value of
    ai.trait.<name of trait here>.default
    is checked.  Again if it is not the sentinel value it will ovverride
    the returned value for trait.

    If trait is not overriden by one of the above values, then the
    default is used.

    Here is an example section providing override values aggression and the
    empire-id trait.

    <ai>
      <trait>
        <aggression>
          <force>
            <enabled>1</enabled>
          </force>
          <default>4</default>
          <ai_0>5</ai_0>
          <ai_1>4</ai_1>
          <ai_2>3</ai_2>
          <ai_3>2</ai_3>
          <ai_4>1</ai_4>
          <ai_5>0</ai_5>
        </aggression>
        <empire-id>
          <force>
            <enabled>1</enabled>
          </force>
          <ai_0>5</ai_0>
          <ai_1>4</ai_1>
          <ai_2>3</ai_2>
          <ai_3>2</ai_3>
          <ai_4>1</ai_4>
          <ai_5>0</ai_5>
        </empire-id>
      </trait>
    </ai>

    :param name: Name of the trait.
    :param default: Default value of the trait.
    :param sentinel: A value indicating no valid value.
    :return: The trait
    """

    force_option = f"ai.trait.{name.lower()}.force.enabled"
    if not fo.getOptionsDBOptionBool(force_option):
        return default

    per_id_option = f"ai.trait.{name.lower()}.{fo.playerName().lower()}"
    all_id_option = f"ai.trait.{name.lower()}.default"

    trait = fo.getOptionsDBOptionInt(per_id_option)
    if trait is None or trait == sentinel:
        trait = fo.getOptionsDBOptionInt(all_id_option)

    if trait is None or trait == sentinel:
        trait = default
    else:
        debug("%s trait bypassed and set to %s for %s", name, repr(trait), fo.playerName())
    return trait



from collections.abc import Mapping
from typing import Union

from common.fo_typing import PlanetId, SpeciesName
from empire.survey_lock import survey_universe_lock
from freeorion_tools.caching import cache_for_current_turn


@survey_universe_lock
def can_build_ship_for_species(species_name: Union[SpeciesName, str]):
    return species_name in get_ship_builders()


@survey_universe_lock
def get_ship_builder_locations(species_name: SpeciesName) -> list[PlanetId]:
    return get_ship_builders().get(species_name, [])


def set_ship_builders(species_name: SpeciesName, pid: PlanetId):
    """
    Add planet where you can build ships for species.

    Warning! Temporal coupling.
    All calls of this function should be done before using of this information.
    """
    get_shipyards().add(pid)
    _get_ship_builders().setdefault(species_name, []).append(pid)


@survey_universe_lock
def get_ship_builders() -> Mapping[SpeciesName, list[PlanetId]]:
    """
    Return map from the species to list of the planet where you could build a ship with it.
    """
    return _get_ship_builders()


@cache_for_current_turn
def _get_ship_builders() -> dict[SpeciesName, list[PlanetId]]:
    """
    Return mutable state.
    """
    return {}


@cache_for_current_turn
def get_shipyards() -> set[PlanetId]:
    return set()


@survey_universe_lock
def has_shipyard(pid: PlanetId) -> bool:
    return pid in get_shipyards()



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_XENTRONIUM_HULL",
    description="SH_XENTRONIUM_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=600 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_SHIP_HULLS_CATEGORY"],
    prerequisites=["SHP_XENTRONIUM_PLATE"],
    unlock=[Item(type=UnlockShipHull, name="SH_XENTRONIUM")],
    graphic="icons/ship_hulls/xentronium_hull_small.png",
)



import freeOrionAIInterface as fo
import random
from itertools import islice
from logging import debug, error, info, warning

import AIDependencies as Dep
import AIstate
import ColonisationAI
import ShipDesignAI
import TechsListsAI
from aistate_interface import get_aistate
from common.print_utils import as_columns
from empire.colony_builders import get_colony_builders
from freeorion_tools import tech_is_complete
from ProductionAI import translators_wanted
from turn_state import (
    get_empire_planets_by_species,
    have_asteroids,
    have_gas_giant,
    have_nest,
    have_ruins,
)


class Choices:
    # Cannot construct on import, because fo.getEmpire() is None at this time
    def init(self):
        rng = random.Random()
        rng.seed(fo.getEmpire().name + fo.getGalaxySetupData().seed)
        self.engine = rng.random() < 0.7
        self.fuel = rng.random() < 0.7
        self.hull = rng.randrange(4)
        self.extra_organic_hull = rng.random() < 0.05
        self.extra_robotic_hull = rng.random() < 0.05
        self.extra_asteroid_hull = rng.random() < 0.05
        self.extra_energy_hull = rng.random() < 0.05


choices = Choices()

# Priorities
ZERO = 0.0
LOW = 0.1
HIGH = 42


# TODO research AI no longer use this method, rename and move this method elsewhere
def get_research_index():
    return get_aistate().character.get_research_index()


def generate_research_orders():
    """Generate research orders."""
    generate_classic_research_orders()


def generate_default_research_order():
    """
    Generate default research orders.
    Add cheapest technology from possible researches
    until current turn point totally spent.
    """

    empire = fo.getEmpire()
    total_rp = empire.resourceProduction(fo.resourceType.research)

    # get all usable researchable techs not already completed or queued for research

    queued_techs = get_research_queue_techs()

    def is_possible(tech_name):
        return all(
            [
                empire.getTechStatus(tech_name) == fo.techStatus.researchable,
                not tech_is_complete(tech_name),
                not exclude_tech(tech_name),
                tech_name not in queued_techs,
            ]
        )

    # (cost, name) for all tech that possible to add to queue, cheapest last
    possible = sorted(
        [(fo.getTech(tech).researchCost(empire.empireID), tech) for tech in fo.techs() if is_possible(tech)],
        reverse=True,
    )

    debug("Techs in possible list after enqueues to Research Queue:")
    for _, tech in possible:
        debug("    " + tech)
    debug("")

    # iterate through techs in order of cost
    fo.updateResearchQueue()
    total_spent = fo.getEmpire().researchQueue.totalSpent
    debug("Enqueuing techs. already spent RP: %s total RP: %s", total_spent, total_rp)

    while total_rp > 0 and possible:
        cost, name = possible.pop()  # get chipest
        total_rp -= cost
        fo.issueEnqueueTechOrder(name, -1)
        debug("    enqueued tech %s  : cost: %s RP", name, cost)
    debug("")


def get_completed_techs() -> list[str]:
    """Get completed and available for use techs."""
    return [tech for tech in fo.techs() if tech_is_complete(tech)]


def get_research_queue_techs():
    """Get list of techs in research queue."""
    return [element.tech for element in fo.getEmpire().researchQueue]


def exclude_tech(tech_name):
    return (
        (not get_aistate().character.may_research_tech(tech_name))
        or tech_name in TechsListsAI.unusable_techs()
        or tech_name in Dep.UNRESEARCHABLE_TECHS
    )


def research_now(tech_name: str, with_prerequisites: bool = True) -> None:
    todo = [tech_name]
    empire = fo.getEmpire()
    if with_prerequisites:
        tech = fo.getTech(tech_name)
        todo += [t for t in tech.recursivePrerequisites(empire.empireID)]
    for name in todo:
        if not empire.researchQueue.inQueue(name):
            fo.issueEnqueueTechOrder(name, 0)
        else:
            for element in empire.researchQueue:
                if element.tech == name:
                    # research points are not lost when a tech is dequeued
                    fo.issueDequeueTechOrder(name)
                    fo.issueEnqueueTechOrder(name, 0)
                    break


def generate_classic_research_orders():  # noqa: C901
    """generate research orders"""
    empire = fo.getEmpire()
    empire_id = empire.empireID
    aistate = get_aistate()
    enemies_sighted = aistate.misc.get("enemies_sighted", {})
    galaxy_is_sparse = ColonisationAI.galaxy_is_sparse()

    resource_production = empire.resourceProduction(fo.resourceType.research)
    completed_techs = sorted(list(get_completed_techs()))
    _print_research_order_header(resource_production, completed_techs)
    _print_research_queue_head(completed_techs)

    research_queue = empire.researchQueue
    research_queue_list = get_research_queue_techs()
    total_rp = empire.resourceProduction(fo.resourceType.research)
    #
    # set starting techs, or after turn 100 add any additional default techs
    #
    if (fo.currentTurn() <= 2) or ((total_rp - research_queue.totalSpent) > 0):
        research_index = get_research_index()
        if fo.currentTurn() == 1:
            # do only this one on first turn, to facilitate use of a turn-1 savegame for testing of alternate
            # research strategies
            new_tech = ["LRN_PHYS_BRAIN", "GRO_PLANET_ECOL"]
        else:
            new_tech = (
                TechsListsAI.sparse_galaxy_techs(research_index)
                if galaxy_is_sparse
                else TechsListsAI.primary_meta_techs(research_index)
            )
        debug("Empire %s (%d) is selecting research index %d", empire.name, empire_id, research_index)
        # techs_to_enqueue = (set(new_tech)-(set(completed_techs)|set(research_queue_list)))
        techs_to_enqueue = new_tech[:]
        tech_base = set(completed_techs + research_queue_list)
        techs_to_add = []
        for tech in techs_to_enqueue:
            if tech not in tech_base:
                this_tech = fo.getTech(tech)
                if this_tech is None:
                    error("Desired tech '%s' appears to not exist" % tech)
                    continue
                missing_prereqs = [
                    preReq for preReq in this_tech.recursivePrerequisites(empire_id) if preReq not in tech_base
                ]
                techs_to_add.extend(missing_prereqs + [tech])
                tech_base.update(missing_prereqs + [tech])
        cum_cost = 0
        debug("  Enqueued Tech: %20s \t\t %8s \t %s", "Name", "Cost", "CumulativeCost")
        for name in techs_to_add:
            try:
                enqueue_res = fo.issueEnqueueTechOrder(name, -1)
                if enqueue_res == 1:
                    this_tech = fo.getTech(name)
                    this_cost = 0
                    if this_tech:
                        this_cost = this_tech.researchCost(empire_id)
                        cum_cost += this_cost
                    debug("    Enqueued Tech: %20s \t\t %8.0f \t %8.0f", name, this_cost, cum_cost)
                else:
                    warning("    Failed attempt to enqueued Tech: " + name)
            except:  # noqa: E722
                warning("    Failed attempt to enqueued Tech: " + name, exc_info=True)

        debug("\n\nAll techs:")
        debug("=" * 20)
        alltechs = fo.techs()
        info(as_columns(sorted(fo.techs()), columns=3))

        debug("\n\nAll unqueued techs:")
        debug("=" * 20)
        # coveredTechs = new_tech+completed_techs
        as_columns([tn for tn in alltechs if tn not in tech_base], columns=3)
        debug("")

        if fo.currentTurn() == 1:
            return
        research_queue_list = get_research_queue_techs()
        def_techs = TechsListsAI.defense_techs_1()
        for def_tech in def_techs:
            if (
                aistate.character.may_research_tech_classic(def_tech)
                and def_tech not in research_queue_list[:5]
                and not tech_is_complete(def_tech)
            ):
                res = fo.issueEnqueueTechOrder(def_tech, min(3, len(research_queue_list)))
                debug("Empire is very defensive, so attempted to fast-track %s, got result %d", def_tech, res)
    elif fo.currentTurn() > 100:
        generate_default_research_order()

    research_queue_list = get_research_queue_techs()
    num_techs_accelerated = 1  # will ensure leading tech doesn't get dislodged
    got_ggg_tech = tech_is_complete("PRO_ORBITAL_GEN")
    got_sym_bio = tech_is_complete("GRO_SYMBIOTIC_BIO")
    got_xeno_gen = tech_is_complete("GRO_XENO_GENETICS")
    #
    # Consider accelerating techs; priority is
    # Supply/Detect range
    # xeno arch
    # ast / GG
    # gro xeno gen
    # distrib thought
    # quant net
    # pro sing gen
    # death ray 1 cleanup

    nest_tech = Dep.NEST_DOMESTICATION_TECH
    artif_minds = Dep.LRN_ARTIF_MINDS_1

    if have_nest() and not tech_is_complete(nest_tech):
        if artif_minds in research_queue_list:
            insert_idx = 1 + research_queue_list.index(artif_minds)
        else:
            insert_idx = 1
        res = fo.issueEnqueueTechOrder(nest_tech, insert_idx)
        num_techs_accelerated += 1
        debug("Have a monster nest, so attempted to fast-track %s, got result %d", nest_tech, res)
        research_queue_list = get_research_queue_techs()

    #

    #
    # check to accelerate xeno_arch
    if True:  # just to help with cold-folding /  organization
        if (
            have_ruins()
            and not tech_is_complete("LRN_XENOARCH")
            and aistate.character.may_research_tech_classic("LRN_XENOARCH")
        ):
            if artif_minds in research_queue_list:
                insert_idx = 7 + research_queue_list.index(artif_minds)
            elif "GRO_SYMBIOTIC_BIO" in research_queue_list:
                insert_idx = research_queue_list.index("GRO_SYMBIOTIC_BIO") + 1
            else:
                insert_idx = num_techs_accelerated
            if "LRN_XENOARCH" not in research_queue_list[:insert_idx]:
                for xenoTech in ["LRN_XENOARCH", "LRN_TRANSLING_THT", "LRN_PHYS_BRAIN", "LRN_ALGO_ELEGANCE"]:
                    if not tech_is_complete(xenoTech) and xenoTech not in research_queue_list[: (insert_idx + 4)]:
                        res = fo.issueEnqueueTechOrder(xenoTech, insert_idx)
                        num_techs_accelerated += 1
                        debug(
                            "ANCIENT_RUINS: have an ancient ruins, so attempted to fast-track %s to enable LRN_XENOARCH, got result %d",
                            xenoTech,
                            res,
                        )
                research_queue_list = get_research_queue_techs()

    #
    # check to accelerate asteroid or GG tech
    if True:  # just to help with cold-folding / organization
        if have_asteroids():
            insert_idx = (
                num_techs_accelerated
                if "GRO_SYMBIOTIC_BIO" not in research_queue_list
                else research_queue_list.index("GRO_SYMBIOTIC_BIO")
            )
            ast_tech = "PRO_MICROGRAV_MAN"
            if not (tech_is_complete(ast_tech) or ast_tech in research_queue_list[: (1 + insert_idx)]):
                res = fo.issueEnqueueTechOrder(ast_tech, insert_idx)
                num_techs_accelerated += 1
                debug(
                    "Asteroids: plan to colonize an asteroid belt, so attempted to fast-track %s , got result %d",
                    ast_tech,
                    res,
                )
                research_queue_list = get_research_queue_techs()
            elif tech_is_complete("SHP_ZORTRIUM_PLATE"):
                insert_idx = (
                    (1 + insert_idx)
                    if "LRN_FORCE_FIELD" not in research_queue_list
                    else max(1 + insert_idx, research_queue_list.index("LRN_FORCE_FIELD") - 1)
                )
                for ast_tech in ["SHP_ASTEROID_HULLS", "SHP_IMPROVED_ENGINE_COUPLINGS"]:
                    if not tech_is_complete(ast_tech) and ast_tech not in research_queue_list[: insert_idx + 1]:
                        res = fo.issueEnqueueTechOrder(ast_tech, insert_idx)
                        num_techs_accelerated += 1
                        insert_idx += 1
                        debug(
                            "Asteroids: plan to colonize an asteroid belt, so attempted to fast-track %s , got result %d",
                            ast_tech,
                            res,
                        )
                research_queue_list = get_research_queue_techs()
        if have_gas_giant() and not tech_is_complete("PRO_ORBITAL_GEN"):
            fusion_idx = (
                0 if "PRO_FUSION_GEN" not in research_queue_list else (1 + research_queue_list.index("PRO_FUSION_GEN"))
            )
            forcefields_idx = (
                0
                if "LRN_FORCE_FIELD" not in research_queue_list
                else (1 + research_queue_list.index("LRN_FORCE_FIELD"))
            )
            insert_idx = max(fusion_idx, forcefields_idx) if enemies_sighted else fusion_idx
            if "PRO_ORBITAL_GEN" not in research_queue_list[: insert_idx + 1]:
                res = fo.issueEnqueueTechOrder("PRO_ORBITAL_GEN", insert_idx)
                num_techs_accelerated += 1
                debug(
                    "GasGiant: plan to colonize a gas giant, so attempted to fast-track %s, got result %d",
                    "PRO_ORBITAL_GEN",
                    res,
                )
                research_queue_list = get_research_queue_techs()
    #
    # assess if our empire has any non-lousy colonizers, & boost gro_xeno_gen if we don't
    if True:  # just to help with cold-folding / organization
        if got_ggg_tech and got_sym_bio and (not got_xeno_gen):
            most_adequate = 0
            for specName in get_colony_builders():
                environs = {}
                this_spec = fo.getSpecies(specName)
                if not this_spec:
                    continue
                for ptype in [
                    fo.planetType.swamp,
                    fo.planetType.radiated,
                    fo.planetType.toxic,
                    fo.planetType.inferno,
                    fo.planetType.barren,
                    fo.planetType.tundra,
                    fo.planetType.desert,
                    fo.planetType.terran,
                    fo.planetType.ocean,
                    fo.planetType.asteroids,
                ]:
                    environ = this_spec.getPlanetEnvironment(ptype)
                    environs.setdefault(environ, []).append(ptype)
                most_adequate = max(most_adequate, len(environs.get(fo.planetEnvironment.adequate, [])))
            if most_adequate == 0:
                insert_idx = num_techs_accelerated
                for xg_tech in ["GRO_XENO_GENETICS", "GRO_GENETIC_ENG"]:
                    if (
                        xg_tech not in research_queue_list[: 1 + num_techs_accelerated]
                        and not tech_is_complete(xg_tech)
                        and aistate.character.may_research_tech_classic(xg_tech)
                    ):
                        res = fo.issueEnqueueTechOrder(xg_tech, insert_idx)
                        num_techs_accelerated += 1
                        debug(
                            "Empire has poor colonizers, so attempted to fast-track %s, got result %d",
                            xg_tech,
                            res,
                        )
                research_queue_list = get_research_queue_techs()
    #
    # check to accelerate translinguistics
    if True:  # just to help with cold-folding / organization
        # planet is needed to determine the cost. Without a capital we have bigger problems anyway...
        if not tech_is_complete("LRN_TRANSLING_THT") and translators_wanted():
            insert_idx = num_techs_accelerated
            for dt_ech in ["LRN_TRANSLING_THT", "LRN_PHYS_BRAIN", "LRN_ALGO_ELEGANCE"]:
                if (
                    dt_ech not in research_queue_list[: insert_idx + 2]
                    and not tech_is_complete(dt_ech)
                    and aistate.character.may_research_tech_classic(dt_ech)
                ):
                    res = fo.issueEnqueueTechOrder(dt_ech, insert_idx)
                    num_techs_accelerated += 1
                    insert_idx += 1
                    fmt_str = "Empire wants to build translators, so attempted to fast-track %s (got result %d)"
                    fmt_str += " with current target_RP %.1f and current pop %.1f, on turn %d"
                    debug(fmt_str, dt_ech, res, resource_production, empire.population(), fo.currentTurn())
            research_queue_list = get_research_queue_techs()
    #
    # check to accelerate distrib thought
    if True:  # just to help with cold-folding / organization
        if not tech_is_complete("LRN_DISTRIB_THOUGHT"):
            got_telepathy = False
            for specName in get_empire_planets_by_species():
                this_spec = fo.getSpecies(specName)
                if this_spec and ("TELEPATHIC" in list(this_spec.tags)):
                    got_telepathy = True
                    break
            pop_threshold = 100 if got_telepathy else 300
            if empire.population() > pop_threshold:
                insert_idx = num_techs_accelerated
                for dt_ech in ["LRN_PHYS_BRAIN", "LRN_TRANSLING_THT", "LRN_PSIONICS", "LRN_DISTRIB_THOUGHT"]:
                    if (
                        dt_ech not in research_queue_list[: insert_idx + 2]
                        and not tech_is_complete(dt_ech)
                        and aistate.character.may_research_tech_classic(dt_ech)
                    ):
                        res = fo.issueEnqueueTechOrder(dt_ech, insert_idx)
                        num_techs_accelerated += 1
                        insert_idx += 1
                        fmt_str = "Empire has a telepathic race, so attempted to fast-track %s (got result %d)"
                        fmt_str += " with current target_RP %.1f and current pop %.1f, on turn %d"
                        debug(fmt_str, dt_ech, res, resource_production, empire.population(), fo.currentTurn())
                research_queue_list = get_research_queue_techs()

    # if we own a blackhole, accelerate sing_gen and conc camp
    if True:  # just to help with cold-folding / organization
        if (
            fo.currentTurn() > 50
            and len(AIstate.empireStars.get(fo.starType.blackHole, [])) != 0
            and aistate.character.may_research_tech_classic("PRO_SINGULAR_GEN")
            and not tech_is_complete(Dep.PRO_SINGULAR_GEN)
            and tech_is_complete("LRN_EVERYTHING")
        ):
            # sing_tech_list = [ "LRN_GRAVITONICS" , "PRO_SINGULAR_GEN"]  # formerly also "CON_ARCH_PSYCH", "CON_CONC_CAMP",
            sing_gen_tech = fo.getTech(Dep.PRO_SINGULAR_GEN)
            sing_tech_list = [
                pre_req for pre_req in sing_gen_tech.recursivePrerequisites(empire_id) if not tech_is_complete(pre_req)
            ]
            sing_tech_list += [Dep.PRO_SINGULAR_GEN]
            for singTech in sing_tech_list:
                if singTech not in research_queue_list[: num_techs_accelerated + 1]:
                    res = fo.issueEnqueueTechOrder(singTech, num_techs_accelerated)
                    num_techs_accelerated += 1
                    debug(
                        "have a black hole star outpost/colony, so attempted to fast-track %s, got result %d",
                        singTech,
                        res,
                    )
            research_queue_list = get_research_queue_techs()

    #
    # if got deathray from Ruins, remove most prereqs from queue
    if True:  # just to help with cold-folding / organization
        if tech_is_complete("SHP_WEAPON_4_1"):
            this_tech = fo.getTech("SHP_WEAPON_4_1")
            if this_tech:
                missing_prereqs = [
                    preReq for preReq in this_tech.recursivePrerequisites(empire_id) if preReq in research_queue_list
                ]
                if len(missing_prereqs) > 2:  # leave plasma 4 and 3 if up to them already
                    for preReq in missing_prereqs:  # sorted(missing_prereqs, reverse=True)[2:]
                        if preReq in research_queue_list:
                            fo.issueDequeueTechOrder(preReq)
                    research_queue_list = get_research_queue_techs()
                    if "SHP_WEAPON_4_2" in research_queue_list:  # (should be)
                        idx = research_queue_list.index("SHP_WEAPON_4_2")
                        fo.issueEnqueueTechOrder("SHP_WEAPON_4_2", max(0, idx - 18))

    # TODO: Remove the following example code
    # Example/Test code for the new ShipDesigner functionality
    techs = ["SHP_WEAPON_4_2", "SHP_TRANSSPACE_DRIVE", "SHP_INTSTEL_LOG", "SHP_ASTEROID_HULLS", ""]
    for tech in techs:
        this_tech = fo.getTech(tech)
        if not this_tech:
            debug("Invalid Tech specified")
            continue
        unlocked_items = this_tech.unlockedItems
        unlocked_hulls = []
        unlocked_parts = []
        for item in unlocked_items:
            if item.type == fo.unlockableItemType.shipPart:
                debug("Tech %s unlocks a ShipPart: %s", tech, item.name)
                unlocked_parts.append(item.name)
            elif item.type == fo.unlockableItemType.shipHull:
                debug("Tech %s unlocks a ShipHull: %s", tech, item.name)
                unlocked_hulls.append(item.name)
        if not (unlocked_parts or unlocked_hulls):
            debug("No new ship parts/hulls unlocked by tech %s", tech)
            continue
        old_designs = ShipDesignAI.WarShipDesigner().optimize_design(consider_fleet_count=False)
        new_designs = ShipDesignAI.WarShipDesigner().optimize_design(
            additional_hulls=unlocked_hulls, additional_parts=unlocked_parts, consider_fleet_count=False
        )
        if not (old_designs and new_designs):
            # AI is likely defeated; don't bother with logging error message
            continue
        old_rating, old_pid, old_design_id, old_cost, old_stats = old_designs[0]
        old_design = fo.getShipDesign(old_design_id)
        new_rating, new_pid, new_design_id, new_cost, new_stats = new_designs[0]
        new_design = fo.getShipDesign(new_design_id)
        if new_rating > old_rating:
            debug("Tech %s gives access to a better design!", tech)
            debug("old best design: Rating %.5f", old_rating)
            debug("old design specs: %s - %s", old_design.hull, list(old_design.parts))
            debug("new best design: Rating %.5f", new_rating)
            debug("new design specs: %s - %s", new_design.hull, list(new_design.parts))
        else:
            debug("Tech %s gives access to new parts or hulls but there seems to be no military advantage.", tech)


def _print_research_queue_head(completed_techs: list[str], first_n_techs=10):
    empire_id = fo.getEmpire().empireID
    research_queue = fo.getEmpire().researchQueue

    debug("Techs currently at head of Research Queue")

    for element in islice(research_queue, first_n_techs):
        this_tech = fo.getTech(element.tech)
        if not this_tech:
            warning("Can't retrieve tech %s", element.tech)
            continue
        missing_prereqs = [
            pre_req for pre_req in this_tech.recursivePrerequisites(empire_id) if pre_req not in completed_techs
        ]
        unlocked_items = [uli.name for uli in this_tech.unlockedItems]
        if not missing_prereqs:
            debug(
                "%-27s allocated %6.2f RP (%d turns left) -- unlocks: %s",
                element.tech,
                element.allocation,
                element.turnsLeft,
                unlocked_items,
            )
        else:
            debug(
                "%-27s allocated %6.2f RP -- missing preReqs: %s -- unlocks: %s",
                element.tech,
                element.allocation,
                missing_prereqs,
                unlocked_items,
            )
    debug("")


def _print_research_order_header(resource_production, completed_techs):
    debug("Research Queue Management:")
    debug("\nTotal Current Research Points: %.2f\n", resource_production)
    debug("Techs researched and available for use:")

    tlist = completed_techs + [" "] * 3
    for tline in zip(tlist[0::3], tlist[1::3], tlist[2::3]):
        debug("%25s %25s %25s", *tline)
    debug("")



import copy
from typing import TypeVar

KT = TypeVar("KT")
VT = TypeVar("VT")


def _recursive_dict_diff(
    dict_new: dict[KT, VT], dict_old: dict[KT, VT], dict_diff: dict[KT, VT], diff_level_threshold=0
) -> int:
    NO_DIFF = 9999
    min_diff_level = NO_DIFF
    for key, value in dict_new.items():
        if key not in dict_old:
            dict_diff[key] = copy.deepcopy(value)
            min_diff_level = 0
        elif isinstance(value, dict):
            this_diff_level = (
                _recursive_dict_diff(value, dict_old[key], dict_diff.setdefault(key, {}), diff_level_threshold) + 1
            )
            min_diff_level = min(min_diff_level, this_diff_level)
            if this_diff_level > NO_DIFF and min_diff_level > diff_level_threshold:
                del dict_diff[key]
        elif key not in dict_old or value != dict_old[key]:
            dict_diff[key] = copy.deepcopy(value)
            min_diff_level = 0
    return min_diff_level


def recursive_dict_diff(dict_new: dict[KT, VT], dict_old: dict[KT, VT], diff_level_threshold=0) -> dict[KT, VT]:
    """
    Find the entries in dict_new that are not present in dict_old and store them in dict_diff.

    Example usage:
    dict_a = {1:2, 2: {2: 3, 3: 4}}
    dict_b = {2: {2: 3, 3: 3}}
    recursive_dict_diff(dict_a, dict_b)
    >>> {1:2, 2:{3:4}}

    :param dict_diff: Difference between dict_old and dict_new, modified and filled within this function
    :param diff_level_threshold: Depth to next diff up to which non-diff entries are stored in dict_diff
    :return: recursive depth distance to entries differing in dict_new and dict_old
    """

    diff = {}
    _recursive_dict_diff(dict_new, dict_old, diff, diff_level_threshold=diff_level_threshold)
    return diff
